import numpy as np
from typing import List, Tuple, Union, List

# Contains some typing used in the library
Coordinate = Union[Tuple[Union[float, int], Union[float, int]],
                   List[Union[float, int]]]
Coordinates = List[np.ndarray]

Number = Union[float, int]



import numpy as np
import copy
import time

from pygdsdesign import GdsLibrary, PolygonSet, Rectangle, Text, MicroStrip

start = time.time()

# Chip size of 5mm x 7mm
chip_dx = 5000
chip_dy = 7000

layer = {'nb':1, 'name': 'Al'}

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('TOP')


# Create empty polygon to which other polygon will be added
tot = PolygonSet()

# Chip mark
p = Rectangle((0,0), (200, 10), layer=layer['nb'], name=layer['name']).center()
p += Rectangle((0,0), (10, 200), layer=layer['nb'], name=layer['name']).center()
dx_d, dy_d = p.get_size()
tot += p
tot += copy.copy(p).translate(chip_dx, 0)
tot += copy.copy(p).translate(chip_dx, -chip_dy)
tot += copy.copy(p).translate(0., -chip_dy)


# Create design polygon which will contain one feedline and one resonator
p = PolygonSet()

# Feedline
m = MicroStrip(width=500, layer=layer['nb'], name=layer['name'])
m.add_line(500, 0)
m.add_taper_cosec(100, 0, 10, 10)
m.add_line(chip_dx-(500+100)*2., 0)
m.add_taper_cosec(100, 0, 500, 10)
m.add_line(500, 0)
p += m
dx_f, dy_f = m.get_size()


# Resonator
def spiral(u, args):

    r = args[1]*2.*args[2]*u + args[0]/2.
    theta = args[1]*2.*np.pi*u
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return x, y

def dspiral_dt(u, args):

    r = args[1]*2.*args[2]*u
    theta = args[1]*2.*np.pi*u
    dx_dt = -args[1]*2.*np.pi*np.sin(theta)*r + np.cos(theta)*args[1]*2.*args[2]
    dy_dt =  args[1]*2.*np.pi*np.cos(theta)*r + np.sin(theta)*args[1]*2.*args[2]
    return dx_dt, dy_dt


t = np.linspace(0, 1, 10000)
inner_diameter = 10
nb_turn = 40
spiral_spacing = 3.
spiral_width = 2.

m = MicroStrip(width=spiral_width, layer=layer['nb'], name=layer['name'])
m.add_parametric_curve(spiral, dspiral_dt, t,
                                args=(inner_diameter, nb_turn, spiral_spacing))
dx_s, dy_s = m.get_size()
p += m.translate(dx_f/2.-dx_s/2., dy_s/2.+dy_f/2.+8.5)


# Copy the feedline and the resonator 4 times
p += copy.copy(p).translate(0, -1500)
p += copy.copy(p).translate(0, -3000)


# Add label on the top right of the chip
t = Text('R0C0', 250, layer=layer['nb'], name=layer['name'])
dx_t, dy_t = t.get_size()
tot += t.translate(-dx_t+chip_dx-250, -dy_t-250)


# Center the design to the chip and the chip to the (0,0) of the cell
tot.center()
tot += p.center()

# Add polygons to cell
cell.add(tot)

# Save GDS file
lib.export_gds('spiral_resonator.gds')

stop = time.time()

print('Generation time:', stop-start, ' s')



import time

from pygdsdesign import GdsLibrary, PolygonSet, shapes

# Layer definitions (0 is reserved for default)
layers_crossover = {'layer_dielectric' : 2,
                    'name_dielectric' : 'SiO2',
                    'layer_metal' : 3,
                    'name_metal' : 'Al'}
layers_daisychain = {'layer_NbN_etch' : 1,
                     'name_NbN_etch' : 'NbN',
                     'layer_dielectric' : 2,
                     'name_dielectric' : 'SiO2',
                     'layer_metal' : 3,
                     'name_metal' : 'Al'}

# some dimensions
gap = 50
pad_size = 200

# Start the design

start = time.time()

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('FIRST')

# Create empty polygon to which other polygon will be added
tot = PolygonSet()

for i in range(20):
    tot += shapes.crossover(**layers_crossover).translate(150*i, 0)

tot.center()

# Add some daisychains to test
# a geometry for briding our feedlines
tot += shapes.daisychain(num=1, l=160, w=5, **layers_daisychain).translate(0, 3000)
tot += shapes.daisychain(num=2, l=20, w=50, **layers_daisychain).translate(0, 2500)
tot += shapes.daisychain(num=4, l=10, w=1, **layers_daisychain).translate(0, 2000)
tot += shapes.daisychain(num=8, l=5, w=5, **layers_daisychain).translate(0, 1500)
tot += shapes.daisychain(num=16, l=100, w=1, **layers_daisychain).translate(0, 1000)
# add one with undercut for the dielectric layer on different datatype
tot += shapes.daisychain(num=16, l=100, w=1, u=0.5, **layers_daisychain).translate(0, 1000)


# Add polygons to cell
cell.add(tot)

lib.export_gds('crossover.gds')

stop = time.time()

print('Generation time:', stop-start, ' s')


from pygdsdesign import *
import json
import os
import pytest

PATH = os.path.dirname(os.path.realpath(__file__))


with open(os.path.join(PATH, 'polygons.json')) as f:
    polygons = json.load(f)


with open(os.path.join(PATH, 'results.json')) as f:
    results = json.load(f)


@pytest.fixture
def square():
    return PolygonSet(polygons['square']['points'])


@pytest.fixture
def cross():
    return PolygonSet(polygons['cross']['points'])


@pytest.fixture
def triangle():
    return PolygonSet(polygons['triangle']['points'])


@pytest.mark.usefixtures("square", "cross", "triangle")
def test_translate(square,cross,triangle):

    #Check a basic translate
    result = square.translate(10,10)
    assert np.isclose(np.asarray([[[0,0],[0,20],[20,20],[20,0]]]), np.asarray(result.polygons)).all()

    #Check a translate of 0
    result = square.translate(0,0)
    assert np.isclose(np.asarray(square.polygons), np.asarray(result.polygons)).all()

    #Check huge translate
    result = cross.translate(9999999999,9999999999)
    assert np.isclose(np.asarray([[[9.99999998e+09,1.00000000e+10],[9.99999998e+09,1.00000000e+10],[1.00000000e+10,1.00000000e+10],[1.00000000e+10,1.00000000e+10]],[[1.00000000e+10,9.99999998e+09],[1.00000000e+10,1.00000000e+10],[1.00000000e+10,1.00000000e+10],[1.00000000e+10,9.99999998e+09]]]), np.asarray(result.polygons)).all()


@pytest.mark.usefixtures("square", "cross", "triangle")
def test_crop(square,cross,triangle):

    #Check a basic crop
    result = crop(cross,'top',10)
    assert np.isclose(np.asarray([[[2.5,-2.5],[17.5,-2.5],[17.5,2.5],[2.5,2.5],[2.5,7.5],[-2.5,7.5],[-2.5,2.5],[-17.5,2.5],[-17.5,-2.5],[-2.5,-2.5],[-2.5,-17.5],[2.5,-17.5]]]), np.asarray(result.polygons)).all()


@pytest.mark.usefixtures("square", "cross", "triangle")
def test_flip(square,cross,triangle):

    #Check x flip
    cross_copy = PolygonSet(cross.polygons)
    result = cross_copy.flip('x')
    assert np.isclose(np.asarray([[[-17.5,2.5],[-17.5,-2.5],[17.5,-2.5],[17.5,2.5]],[[-2.5,17.5],[-2.5,-17.5],[2.5,-17.5],[2.5,17.5]]]), np.asarray(result.polygons)).all()

    #Check y flip
    cross_copy = PolygonSet(cross.polygons)
    result = cross_copy.flip('y')
    assert np.isclose(np.asarray([[[17.5,-2.5],[17.5,2.5],[-17.5,2.5],[-17.5,-2.5]],[[2.5,-17.5],[2.5,17.5],[-2.5,17.5],[-2.5,-17.5]]]), np.asarray(result.polygons)).all()


@pytest.mark.usefixtures("square", "cross", "triangle")
def test_center(square,cross,triangle):

    #Check x flip
    poly = PolygonSet([[(1,2),(3,4)]])
    result = poly.center()
    assert np.isclose(np.asarray([[[-1., -1.],[ 1.,  1.]]]), np.asarray(result.polygons)).all()


@pytest.mark.usefixtures("square", "cross", "triangle")
def test_get_center(square,cross,triangle):

    #Check x flip
    poly = PolygonSet([[(1,2),(3,4)]])
    result = poly.get_center()
    assert result == (2.0, 3.0)



@pytest.mark.usefixtures("square", "cross", "triangle")
def test_get_bounding_box(square,cross,triangle):

    result = triangle.get_bounding_box()
    assert np.isclose(np.asarray([[-10, -10],[ 10,  10]]), np.asarray(result)).all()


@pytest.mark.usefixtures("square", "cross", "triangle")
def test_get_area(square,cross,triangle):

    result = cross.get_area()
    assert 350.0 == result

@pytest.mark.usefixtures("square", "cross", "triangle")
def test_rotate(square,cross,triangle):

    result = cross.rotate(99)
    assert np.isclose(np.asarray([[[-3.19488249,17.3865674],[1.80115168,17.5856718],[3.19488249,-17.3865674],[-1.80115168,-17.5856718]],[[-17.5856718,1.80115168],[17.3865674,3.19488249],[17.5856718,-1.80115168],[-17.3865674,-3.19488249]]]), np.asarray(result.polygons)).all()


@pytest.mark.usefixtures("square", "cross", "triangle")
def test_scale(square,cross,triangle):

    result = cross.scale(99,99)
    assert np.isclose(np.asarray([[[-1732.5,-247.5],[-1732.5,247.5],[1732.5,247.5],[1732.5,-247.5]],[[-247.5,-1732.5],[-247.5,1732.5],[247.5,1732.5],[247.5,-1732.5]]]), np.asarray(result.polygons)).all()

@pytest.mark.usefixtures("square", "cross", "triangle")
def test_scale(square,cross,triangle):

    result = cross.mirror((-10,10), (10,-10))
    assert np.isclose(np.asarray([[[2.5,17.5],[-2.5,17.5],[-2.5,-17.5],[2.5,-17.5]],[[17.5,2.5],[-17.5,2.5],[-17.5,-2.5],[17.5,-2.5]]]), np.asarray(result.polygons)).all()


@pytest.mark.usefixtures("square", "cross", "triangle")
def test_change_layer(square,cross,triangle):

    #Check a basic change_layer
    poly = PolygonSet(triangle.polygons, [1],[2],["test"],["blue"])
    result = poly.change_layer(2,3,"sucess1","sucess2")
    assert np.isclose(np.asarray([[[-10,-10],[0,10],[10,-10]]]), np.asarray(result.polygons)).all()
    assert result.layers==[2]
    assert result.datatypes==[3]
    assert result.colors==["sucess1"]
    assert result.names==["sucess2"]

    #Check a change_layer on a empty polygon
    poly = PolygonSet()
    result = poly.change_layer(2,3,"sucess1","sucess2")
    assert np.isclose(np.asarray([[0,0]]), np.asarray(result.polygons)).all()
    assert result.layers==[2]
    assert result.datatypes==[3]
    assert result.colors==["sucess1"]
    assert result.names==["sucess2"]


# TODO : correct this test
# @pytest.mark.usefixtures("square", "cross", "triangle")
# def test_remove_polygon_from_layer(square,cross,triangle):

#     poly1 = PolygonSet(triangle.polygons, [1],[2],["test"],["blue"])
#     poly2 = PolygonSet(triangle.polygons, [2],[2],["test"],["blue"])
#     poly3 = PolygonSet(triangle.polygons, [2],[2],["test"],["blue"])
#     poly1 += poly2
#     poly1 += poly3
#     result = poly1.remove_polygon_from_layer(2)
#     assert np.isclose(np.asarray([[[-10,-10],[0,10],[10,-10]]]), np.asarray(result.polygons[0])).all()



import warnings
import struct
import numpy
import hashlib


def _record_reader(stream):
    """
    Generator for complete records from a GDSII stream file.

    Parameters
    ----------
    stream : file
        GDSII stream file to be read.

    Returns
    -------
    out : list
        Record type and data (as a numpy.array, string or None)
    """
    while True:
        header = stream.read(4)
        if len(header) < 4:
            return
        size, rec_type = struct.unpack(">HH", header)
        data_type = rec_type & 0x00FF
        rec_type = rec_type // 256
        data = None
        if size > 4:
            if data_type == 0x01:
                data = numpy.array(
                    struct.unpack(
                        ">{0}H".format((size - 4) // 2), stream.read(size - 4)
                    ),
                    dtype="uint",
                )
            elif data_type == 0x02:
                data = numpy.array(
                    struct.unpack(
                        ">{0}h".format((size - 4) // 2), stream.read(size - 4)
                    ),
                    dtype="int",
                )
            elif data_type == 0x03:
                data = numpy.array(
                    struct.unpack(
                        ">{0}l".format((size - 4) // 4), stream.read(size - 4)
                    ),
                    dtype="int",
                )
            elif data_type == 0x05:
                data = numpy.array(
                    [
                        _eight_byte_real_to_float(stream.read(8))
                        for _ in range((size - 4) // 8)
                    ]
                )
            else:
                data = stream.read(size - 4)
                if str is not bytes:
                    if data[-1] == 0:
                        data = data[:-1].decode("ascii")
                    else:
                        data = data.decode("ascii")
                elif data[-1] == "\0":
                    data = data[:-1]
        yield [rec_type, data]


def _raw_record_reader(stream):
    """
    Generator for complete records from a GDSII stream file.

    Parameters
    ----------
    stream : file
        GDSII stream file to be read.

    Returns
    -------
    out : 2-tuple
        Record type and binary data (including header)
    """
    while True:
        header = stream.read(4)
        if len(header) < 4:
            return
        size, rec_type = struct.unpack(">HH", header)
        rec_type = rec_type // 256
        yield (rec_type, header + stream.read(size - 4))


def _eight_byte_real(value):
    """
    Convert a number into the GDSII 8 byte real format.

    Parameters
    ----------
    value : number
        The number to be converted.

    Returns
    -------
    out : string
        The GDSII binary string that represents `value`.
    """
    if value == 0:
        return b"\x00\x00\x00\x00\x00\x00\x00\x00"
    if value < 0:
        byte1 = 0x80
        value = -value
    else:
        byte1 = 0x00
    fexp = numpy.log2(value) / 4
    exponent = int(numpy.ceil(fexp))
    if fexp == exponent:
        exponent += 1
    mantissa = int(value * 16.0 ** (14 - exponent))
    byte1 += exponent + 64
    byte2 = mantissa // 281474976710656
    short3 = (mantissa % 281474976710656) // 4294967296
    long4 = mantissa % 4294967296
    return struct.pack(">HHL", byte1 * 256 + byte2, short3, long4)


def _eight_byte_real_to_float(value):
    """
    Convert a number from GDSII 8 byte real format to float.

    Parameters
    ----------
    value : string
        The GDSII binary string representation of the number.

    Returns
    -------
    out : float
        The number represented by `value`.
    """
    short1, short2, long3 = struct.unpack(">HHL", value)
    exponent = (short1 & 0x7F00) // 256 - 64
    mantissa = (
        ((short1 & 0x00FF) * 65536 + short2) * 4294967296 + long3
    ) / 72057594037927936.0
    if short1 & 0x8000:
        return -mantissa * 16.0 ** exponent
    return mantissa * 16.0 ** exponent


def gdsii_hash(filename, engine=None):
    """
    Calculate the a hash value for a GDSII file.

    The hash is generated based only on the contents of the cells in the
    GDSII library, ignoring any timestamp records present in the file
    structure.

    Parameters
    ----------
    filename : string or Path
        Path to the GDSII file.
    engine : hashlib-like engine
        The engine that executes the hashing algorithm.  It must provide
        the methods `update` and `hexdigest` as defined in the hashlib
        module.  If None, the dafault `hashlib.sha1()` is used.

    Returns
    -------
    out : string
        The hash correponding to the library contents in hex format.
    """
    if hasattr(filename, "__fspath__"):
        filename = filename.__fspath__()
    with open(filename, "rb") as fin:
        data = fin.read()
    contents = []
    start = pos = 0
    while pos < len(data):
        size, rec = struct.unpack(">HH", data[pos : pos + 4])
        if rec == 0x0502:
            start = pos + 28
        elif rec == 0x0700:
            contents.append(data[start:pos])
        elif size == 0:
            warnings.warn(
                "[pygdsdesign] Zero-length record found in {0} at position {1}. "
                "Skipping the remaining of the file contents.".format(filename, pos),
                stacklevel=2,
            )
            break
        pos += size
    h = hashlib.sha1() if engine is None else engine
    for x in sorted(contents):
        h.update(x)
    return h.hexdigest()


def set_gdsii_timestamp(filename, timestamp):
    """
    Set all timestamps in a given GDSII file.

    The GDSII format includes creation timestamps for the whole library
    and each cell in the contents of the GDSII file.  Those timestamps
    are overwritten by the value passed to this function, which is
    useful for creating GDSII files with identical binary contents,
    i.e., for regression testing.

    Parameters
    ----------
    filename : string or Path
        Path to the GDSII file.
    timestamp : datetime
        The new timestamp to set in the GDSII file binary content.

    Notes
    -----
    This function modifies the *contents* of the file.  The creation,
    modification and access timestamps of the file itself are modified
    according to the file system rules in place for reading and
    writing.

    """
    ts = struct.pack(
        ">12h",
        timestamp.year,
        timestamp.month,
        timestamp.day,
        timestamp.hour,
        timestamp.minute,
        timestamp.second,
        timestamp.year,
        timestamp.month,
        timestamp.day,
        timestamp.hour,
        timestamp.minute,
        timestamp.second,
    )
    # For compatibility
    if hasattr(filename, "__fspath__"):
        filename = filename.__fspath__()
    with open(filename, "r+b") as fio:
        while True:
            data = fio.read(4)
            if len(data) < 4:
                if len(data) > 0:
                    warnings.warn("Possibly corrupted file", stacklevel=2)
                return
            size, rec_type = struct.unpack(">HH", data)
            if rec_type == 0x0102 or rec_type == 0x0502:
                assert size == 28
                fio.write(ts)
            elif size > 4:
                fio.seek(size - 4, 1)



import time

import numpy as np

from pygdsdesign import GdsLibrary
from pygdsdesign.shapes import qubit_layer_19, qubit_layer_42

start = time.time()

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('FIRST')

tot = qubit_layer_42(layer=52)
tot+= qubit_layer_19(layer=51)

cell.add(tot)

lib.export_gds('leti_masks.gds')

stop = time.time()

print('Generation time:', stop-start, ' s')


from functools import partial
import numpy as np
import copy
from typing import Callable, Tuple, Optional, Dict, Union
from typing_extensions import Literal
from scipy.integrate import quad

from pygdsdesign.polygons import Rectangle
from pygdsdesign.polygonSet import PolygonSet
from pygdsdesign.transmission_lines.transmission_line import TransmissionLine
from pygdsdesign.typing_local import Coordinate


class CPW(TransmissionLine):

    def __init__(
        self,
        width: float,
        gap: float,
        layer: int = 0,
        datatype: int = 0,
        name: str = "",
        color: str = "",
        ref: Optional[Coordinate] = None,
    ) -> None:
        """
        Coplanar allows to easily draw a continuous coplanar waveguide.

        Parameters
        ----------
        width : float
            Width of the central line in um
            This width can be modified latter along the strip or smoothly
            by using tappered functions.
        gap : float
            Width of the central line in um
            This width can be modified latter along the strip or smoothly
            by using tappered functions.
        layer : int
            Layer number of the coplanar.
        datatype : int
            Datatype number of the coplanar.
        """

        TransmissionLine.__init__(self, layer=layer,
                                        datatype=datatype,
                                        name=name,
                                        color=color,
                                        ref=ref)

        self._w = width
        self._s = gap
        self._bounding_polygon = PolygonSet()


    @property
    def width(self):
        return self._w


    @width.setter
    def width(self, width:float):
        self._w = width


    @property
    def gap(self):
        return self._s


    @gap.setter
    def gap(self, gap:float):
        self._s = gap


    @property
    def bounding_polygon(self):
        return self._bounding_polygon


    # def translate(self, dx: float,
    #                     dy: float) -> PolygonSet:
    #     """
    #     Translate the cpw by the amount dx, dy in the x and y direction.
    #     Take care of translating also the bounding_polygon.

    #     Args:
    #         dx (float): amount of translation in the x direction in um.
    #         dy (float): amount of translation in the y direction in um.
    #     """

    #     self._bounding_polygon.translate(dx,dy)
    #     return super().translate(dx, dy)


    ###########################################################################
    #
    #                   Add polygons to the existing coplanar waveguide
    #
    ###########################################################################


    def add_line(self, x: float,
                       y: float) -> PolygonSet:
        """
        Add a piece of linear coplanar in the x or y direction .

        Parameters
        ----------
        x : float
            Length of the strip in the x direction in um.
        y : float
            Length of the strip in the y direction in um.
        """
        p  = PolygonSet([[(0, -self._w/2.),
                          (0, -self._w/2. - self._s),
                          (x+y, -self._w/2. - self._s),
                          (x+y, -self._w/2.)]],
                         layers=[self._layer],
                         datatypes=[self._datatype],
                         names=[self._name],
                         colors=[self._color])

        p += PolygonSet([[(0, +self._w/2.),
                            (0, +self._w/2. + self._s),
                            (x+y, +self._w/2. + self._s),
                            (x+y, +self._w/2.)]],
                        layers=[self._layer],
                        datatypes=[self._datatype],
                        names=[self._name],
                        colors=[self._color])

        if y==0 and x>0:
            pass
        elif y==0 and x<0:
            pass
        elif y>0 and x==0:
            p.rotate(np.pi/2.)
        elif y<0 and x==0:
            p.rotate(np.pi/2.)
        else:
            raise ValueError("x and y can't be both != 0")

        self._add(p.translate(*self.ref))
        self.ref = [self.ref[0]+x, self.ref[1]+y]
        self.total_length += abs(x) + abs(y)

        # update bounding polygon
        a,b = p.get_bounding_box()
        bp = Rectangle((a[0], a[1]),
                        (b[0], b[1]))
        self._bounding_polygon+=bp

        return self


    ###########################################################################
    #
    #                       Add turn
    #
    ###########################################################################


    def add_turn(self, radius: float,
                       orientation: Literal['lb', 'l-b', 'left-bottom', 'leftbottom',
                                            'lt', 'l-t', 'left-top', 'lefttop'
                                            'rb', 'r-b', 'right-bottom', 'rightbottom'
                                            'rt', 'r-t', 'right-top', 'righttop'
                                            'tl', 't-l', 'top-left', 'topleft'
                                            'tr', 't-r', 'top-right', 'topright'
                                            'bl', 'b-l', 'bottom-left', 'bottomleft'
                                            'br', 'b-r', 'bottom-right', 'bottomright'
                                            'lt', 'l-t', 'lefttop', 'left-top'],
                       nb_points: int=50) -> PolygonSet:
        """
        Add a circulare turn to the strip.

        Parameters
        ----------
        radius : float
            Radius of the arc in um.
        orientation : str
            Orientation of the turn.
            The logic of the orientation naming is the following, the first term
            refers to your departure point while the second one refers to your
            arrival point, they are so the reverse to one another.
            Example: lt _|.
            Also the orientation can be writen in various way, for example all
            following are equivalent:
            'lb', 'l-b', 'left-bottom', 'leftbottom'
        nb_point : int (default=50)
            Number of point used in the polygon.
        """


        theta = np.linspace(np.pi/2., 0, nb_points)

        x = np.concatenate(((radius+self._w/2.)*np.cos(theta), (radius+self._w/2.+self._s)*np.cos(theta[::-1])))
        y = np.concatenate(((radius+self._w/2.)*np.sin(theta), (radius+self._w/2.+self._s)*np.sin(theta[::-1])))

        p = PolygonSet(polygons=[np.vstack((x, y)).T],
                       layers=[self._layer],
                       datatypes=[self._datatype],
                       names=[self._name],
                       colors=[self._color])

        x = np.concatenate(((radius-self._w/2.)*np.cos(theta), (radius-self._w/2.-self._s)*np.cos(theta[::-1])))
        y = np.concatenate(((radius-self._w/2.)*np.sin(theta), (radius-self._w/2.-self._s)*np.sin(theta[::-1])))

        p += PolygonSet(polygons=[np.vstack((x, y)).T],
                        layers=[self._layer],
                        datatypes=[self._datatype],
                        names=[self._name],
                        colors=[self._color])

        # generate bounding polygon
        x = np.concatenate(((radius+self._w/2.+self._s)*np.cos(theta), (radius-self._w/2.-self._s)*np.cos(theta[::-1])))
        y = np.concatenate(((radius+self._w/2.+self._s)*np.sin(theta), (radius-self._w/2.-self._s)*np.sin(theta[::-1])))

        bp = PolygonSet(polygons=[np.vstack((x, y)).T])

        if orientation.lower() in ['lb', 'l-b', 'left-bottom', 'leftbottom']:
            p.translate(self.ref[0], self.ref[1]-radius)
            bp.translate(self.ref[0], self.ref[1]-radius)
            self.ref = [self.ref[0] + radius, self.ref[1] - radius]

        elif orientation.lower() in ['lt', 'l-t', 'left-top', 'lefttop']:
            p.rotate(-np.pi/2.).translate(self.ref[0], self.ref[1]+radius)
            bp.rotate(-np.pi/2.).translate(self.ref[0], self.ref[1]+radius)
            self.ref = [self.ref[0] + radius, self.ref[1] + radius]

        elif orientation.lower() in ['rb', 'r-b', 'right-bottom', 'rightbottom']:
            p.rotate(-3.*np.pi/2.).translate(self.ref[0], self.ref[1]-radius)
            bp.rotate(-3.*np.pi/2.).translate(self.ref[0], self.ref[1]-radius)
            self.ref = [self.ref[0] - radius, self.ref[1] - radius]

        elif orientation.lower() in ['rt', 'r-t', 'right-top', 'righttop']:
            p.rotate(np.pi).translate(self.ref[0], self.ref[1]+radius)
            bp.rotate(np.pi).translate(self.ref[0], self.ref[1]+radius)
            self.ref = [self.ref[0] - radius, self.ref[1] + radius]

        elif orientation.lower() in ['tl', 't-l', 'top-left', 'topleft']:
            p.translate(-radius+self.ref[0], -radius-self._s-self._w/2.+self.ref[1]).flip('x')
            bp.translate(-radius+self.ref[0], -radius-self._s-self._w/2.+self.ref[1]).flip('x')
            self.ref = [self.ref[0] - radius, self.ref[1] - radius]

        elif orientation.lower() in ['tr', 't-r', 'top-right', 'topright']:
            p.rotate(np.pi).translate(self.ref[0]+radius, self.ref[1])
            bp.rotate(np.pi).translate(self.ref[0]+radius, self.ref[1])
            self.ref = [self.ref[0] + radius, self.ref[1] - radius]

        elif orientation.lower() in ['bl', 'b-l', 'bottom-left', 'bottomleft']:
            p.translate(self.ref[0]-radius, self.ref[1])
            bp.translate(self.ref[0]-radius, self.ref[1])
            self.ref = [self.ref[0] - radius, self.ref[1] + radius]

        elif orientation.lower() in ['br', 'b-r', 'bottom-right', 'bottomright']:
            p.rotate(np.pi/2.).translate(self.ref[0]+radius, self.ref[1])
            bp.rotate(np.pi/2.).translate(self.ref[0]+radius, self.ref[1])
            self.ref = [self.ref[0] + radius, self.ref[1] + radius]
        else:
            raise ValueError('Your orientation "'+orientation+'" is incorrect.')

        self._add(p)
        self._bounding_polygon+=bp
        self.total_length += radius*np.pi/2.

        return self


    def add_fresnel_turn(self, radius: float,
                               orientation: Literal['lb', 'l-b', 'left-bottom', 'leftbottom',
                                                    'lt', 'l-t', 'left-top', 'lefttop'
                                                    'rb', 'r-b', 'right-bottom', 'rightbottom'
                                                    'rt', 'r-t', 'right-top', 'righttop'
                                                    'tl', 't-l', 'top-left', 'topleft'
                                                    'tr', 't-r', 'top-right', 'topright'
                                                    'bl', 'b-l', 'bottom-left', 'bottomleft'
                                                    'br', 'b-r', 'bottom-right', 'bottomright'
                                                    'lt', 'l-t', 'lefttop', 'left-top'],
                               nb_points: int=101) -> PolygonSet:
        """
        Return the parametric Fresnel curve along the length t.
        The curve is doubled, mirrored, joint and normalized to a given radius

        Parameters
        ----------
        radius : float
            Radius of the arc in um.
        orientation : str
            Orientation of the turn.
            The logic of the orientation naming is the following, the first term
            refers to your departure point while the second one refers to your
            arrival point, they are so the reverse to one another.
            Example: lt _|.
            Also the orientation can be writen in various way, for example all
            following are equivalent:
            'lb', 'l-b', 'left-bottom', 'leftbottom'
        nb_points : int (default=101)
            Number of point used in the polygon.
            Must be an odd number
        """



        # All the computations are stored in self.calc to cache the result
        dx, dy, t, x, y, f_dx, f_dy, length = self.calc(radius, nb_points)

        # Prepare functions for the parametric generation
        # Initialize func and dfunc with some values to reuse them later on
        partial_func = partial(self.func, x=x,y=y)
        partial_dfunc = partial(self.dfunc, dx=dx,dy=dy,f_dx=f_dx,f_dy=f_dy)


        # Get the polygon corresponding to a Fresnel turn
        # Add_length is set to False, because of its computational time.
        # We do the computation here and save it for later
        p, bp = self.add_parametric_curve(partial_func,
                                          partial_dfunc,
                                          t,
                                          args=None,
                                          add_polygon=False,
                                          add_length=False)


        # Add the length of the parametric curve only if asked (default)
        if hasattr(self, '_fresnel_lengths'):
            if radius not in self._fresnel_lengths.keys():
                compute_length = True
            else:
                compute_length = False
        else:
            self._fresnel_lengths: Dict[float, float]= {}
            compute_length = True

        if compute_length:
            self._fresnel_lengths[radius] = length

        self.total_length += self._fresnel_lengths[radius]

        # We transform the polygon and move the ref depending on the orientation
        if orientation.lower() in ['lt', 'l-t', 'lefttop', 'left-top']:
            p.translate(0, -self._w/2.)
            bp.translate(0, -self._w/2.)
            self.ref = [p.polygons[0][nb_points][0]+self._w/2., p.polygons[0][nb_points][1]]

        elif orientation.lower() in ['lb', 'l-b', 'leftbottom', 'left-bottom']:
            p.flip('x')
            p.translate(0, -radius+self._s)
            bp.flip('x')
            bp.translate(0, -radius+self._s)
            self.ref = [p.polygons[0][nb_points][0]+self._w/2., p.polygons[0][nb_points][1]]

        elif orientation.lower() in ['br', 'b-r', 'bottomright', 'bottom-right']:
            p.flip('x')
            p.flip('y')
            p.translate(-self._s-self._w/2., self._s)
            bp.flip('x')
            bp.flip('y')
            bp.translate(-self._s-self._w/2., self._s)
            self.ref = [p.polygons[0][-1][0], p.polygons[0][-1][1]+self._w/2.]

        elif orientation.lower() in ['bl', 'b-l', 'bottomright', 'bottom-right']:
            p.rotate(np.pi/2., p.polygons[1][0])
            p.translate(self._s+self._w/2., self._s)
            bp.rotate(np.pi/2., bp.polygons[0][0])
            bp.translate(-(self._s+self._w/2), -(self._s+self._w))
            self.ref = [p.polygons[0][nb_points][0], p.polygons[0][nb_points][1]+self._w/2.]

        elif orientation.lower() in ['tr', 't-r', 'topright', 'top-right']:
            p.rotate(-np.pi/2., p.polygons[1][0])
            p.translate(-self._s-self._w/2., self._s)
            bp.rotate(-np.pi/2., bp.polygons[0][0])
            bp.translate(self._s+self._w/2., -(self._w + self._s))
            self.ref = [p.polygons[0][nb_points][0], p.polygons[0][nb_points][1]-self._w/2.]

        elif orientation.lower() in ['tl', 't-l', 'topleft', 'top-left']:
            p.rotate(-np.pi/2., p.polygons[1][0])
            p.flip('y')
            p.translate(-radius, self._s)
            bp.rotate(-np.pi/2., bp.polygons[0][0])
            bp.flip('y')
            bp.translate(-radius+self._w+2*self._s, -self._w - self._s)
            self.ref = [p.polygons[0][nb_points][0], p.polygons[0][nb_points][1]-self._w/2.]

        elif orientation.lower() in ['rb', 'r-b', 'rightbottom', 'right-bottom']:
            p.rotate(np.pi, p.polygons[1][0])
            p.translate(0,self._s*2.+self._w/2.)
            bp.rotate(np.pi, bp.polygons[0][0])
            bp.translate(0,-(self._s*2.+1.5*self._w))
            self.ref = [p.polygons[0][nb_points][0]-self._w/2., p.polygons[0][nb_points][1]]

        elif orientation.lower() in ['rt', 'r-t', 'righttop', 'right-top']:
            p.flip('y')
            p.translate(-radius-self._s-self._w/2., -self._w/2.)
            bp.flip('y')
            bp.translate(-radius-self._s-self._w/2., -self._w/2.)
            self.ref = [p.polygons[0][nb_points][0]-self._w/2., p.polygons[0][nb_points][1]]

        else:
            raise ValueError('Your orientation "'+orientation+'" is incorrect.')

        self._add(p)
        self._bounding_polygon+=bp
        return self


    ###########################################################################
    #
    #                               Tapers
    #
    ###########################################################################


    def add_taper(self, x: float,
                        y: float,
                        new_width: float,
                        new_gap: float) -> PolygonSet:
        """
        Add linear taper between the current and the new width.

        Parameters
        ----------
        x : float
            Length of the taper in the x direction in um.
        y : float
            Length of the taper in the y direction in um.
        new_width : float
            New width of the coplanar waveguide in um.
        new_gape : float
            New gape of the coplanar waveguide in um.
        """

        p = PolygonSet(polygons=[[(0., self._w/2.)]],
                       layers=[self._layer],
                       datatypes=[self._datatype],
                       names=[self._name],
                       colors=[self._color])
        p>(x+y, new_width/2.-self._w/2.)
        p>(0., new_gap)
        p>(-x-y, -new_gap-new_width/2.+self._s+self._w/2.)
        p += copy.copy(p).mirror((0, 0), (1, 0))

        # generate bounding polygon
        bp = PolygonSet(polygons=[[(0, self._w/2+self._s),
                                  (x+y, new_width/2+new_gap),
                                  (x+y, 0),
                                  (0, 0)]]
                          )
        bp += copy.copy(bp).mirror((0, 0), (1, 0))

        if x>0 and y==0:
            p.translate(*self.ref)
            bp.translate(*self.ref)
        elif x<0 and y==0:
            p.translate(*self.ref)
            bp.translate(*self.ref)
        elif x==0 and y<0:
            p.rotate(np.pi/2.).translate(*self.ref)
            bp.rotate(np.pi/2.).translate(*self.ref)
        elif x==0 and y>0:
            p.rotate(np.pi/2.).translate(*self.ref)
            bp.rotate(np.pi/2.).translate(*self.ref)
        else:
            raise ValueError('x and y should not be both != than 0.')


        self.ref = [self.ref[0]+x, self.ref[1]+y]

        self._add(p)
        self._bounding_polygon+=bp

        self.total_length += abs(x) + abs(y)
        self._w = new_width
        self._s = new_gap

        return self


    def add_parametric_curve(self,
                             f: Callable[..., Tuple[np.ndarray, np.ndarray]],
                             df: Callable[..., Tuple[np.ndarray, np.ndarray]],
                             t: np.ndarray,
                             args: Optional[Tuple[Optional[float], ...]]=None,
                             add_polygon: bool=True,
                             add_length: bool=True) -> Union[PolygonSet, Tuple[PolygonSet, PolygonSet]]:
        """
        Create a coplanar line following the parametric equation f and its
        derivative df along the length t.
        In order to return the curve length correctly, the derivative df of f
        must be correct, its absolute amplitude must be correct.

        Parameters
        ---------
        f : func
            Function calculating the parametric equation.
            Must be of the type f(t, args) and return a tuple of coordinate
            (x, y).
        df : func
            Function calculating the derivative of the parametric equation.
            Must be of the type df(t, args) and return a tuple of coordinate
            (dx, dy).
        t : np.ndarray
            Array of the length of the parametric curve.
            Also determine the number of point of the total polygon.
            Must not necessarily starts at 0.
        args : variable arguments (default None)
            Argument passed to f and df
        """
        if args is None:
            args = (None, )

        dx1, dy1 = df(t, args)
        n = np.hypot(dx1, dy1)
        dx1, dy1 = dx1/n, dy1/n

        x1, y1 = f(t, args)
        theta1 = np.angle(dx1+1j*dy1)-np.pi/2.
        x1, y1 = x1+np.cos(theta1)*(self._w/2.+self._s), y1+np.sin(theta1)*(self._w/2.+self._s)

        dx2, dy2 = df(t[::-1], args)
        n = np.hypot(dx2, dy2)
        dx2, dy2 = dx2/n, dy2/n
        x2, y2 = f(t[::-1], args)
        theta2 = np.angle(dx2+1j*dy2)-np.pi/2.
        x2, y2 = x2+np.cos(theta2)*self._w/2., y2+np.sin(theta2)*self._w/2.

        x, y = np.concatenate((x1, x2)), np.concatenate((y1, y2))
        p1 = np.vstack((x, y)).T
        # keep the coordinates of the outer trace
        bp_x, bp_y = x1, y1

        dx1, dy1 = df(t, args)
        n = np.hypot(dx1, dy1)
        dx1, dy1 = dx1/n, dy1/n

        x1, y1 = f(t, args)
        theta1 = np.angle(dx1+1j*dy1)-np.pi/2.
        x1, y1 = x1+np.cos(theta1)*-(self._w/2.+self._s), y1+np.sin(theta1)*-(self._w/2.+self._s)


        dx2, dy2 = df(t[::-1], args)
        n = np.hypot(dx2, dy2)
        dx2, dy2 = dx2/n, dy2/n
        x2, y2 = f(t[::-1], args)
        theta2 = np.angle(dx2+1j*dy2)-np.pi/2.
        x2, y2 = x2+np.cos(theta2)*-self._w/2., y2+np.sin(theta2)*-self._w/2.

        x, y = np.concatenate((x1, x2)), np.concatenate((y1, y2))
        p2 = np.vstack((x, y)).T
        # add the coordinates of the outer trace in reversed order
        bp_x, bp_y = np.concatenate((bp_x, np.flip(x1))), np.concatenate((bp_y, np.flip(y1)))
        bp = np.vstack((bp_x, bp_y)).T

        p = PolygonSet(polygons=[p1],
                       layers=[self._layer],
                       datatypes=[self._datatype],
                       names=[self._name],
                       colors=[self._color])\
           + PolygonSet(polygons=[p2],
                        layers=[self._layer],
                        datatypes=[self._datatype],
                        names=[self._name],
                        colors=[self._color])
        p = p.translate(self.ref[0], self.ref[1])

        # generate bounding_polygon
        bp = PolygonSet(polygons=[bp])
        bp = bp.translate(self.ref[0], self.ref[1])

        # Calculate curve length
        def func(t, args):
            dx1, dy1 = df(t, args)
            return np.hypot(dx1, dy1)

        # Add the length of the parametric curve only if asked (default)
        if add_length:
            self.total_length += quad(func, t[0], t[-1], args=(args,))[0]

        # Add polygon only if asked (default)
        if add_polygon:
            self.ref = [p.polygons[0][int(len(p.polygons[0])/2)][0]-self._w/2.,
                        p.polygons[0][int(len(p.polygons[0])/2)][1]]
            self._add(p)
            self._bounding_polygon+=bp

            return self
        else:
            return p, bp


    def add_end(self, x: float,
                      y: float,
                      update_ref: bool=False) -> PolygonSet:
        """
        Add an open end to a coplanar waveguide in the x, y direction
        Notice that this method does not increment the total length of the
        waveguide nor its reference point (unless update_ref is set to True)!

        Parameters
        ----------
        x : float
            Length of the end in the x direction in um.
        y : float
            Length of the end in the y direction in um.
        update_ref : bool
            If True update the reference point of the cpw by including the length
            of the end in x and y direction.
            Useful when starting a cpw by an end_point.
        """

        if x ==0.:
            r = Rectangle((-self._w/2.-self._s, 0),
                          (self._w/2.+self._s, y),
                          layer=self._layer,
                          datatype=self._datatype,
                          name=self._name,
                          color=self._color)
        elif y==0.:
            r = Rectangle((0, -self._w/2.-self._s),
                          (x,  self._w/2.+self._s),
                          layer=self._layer,
                          datatype=self._datatype,
                          name=self._name,
                          color=self._color)
        else:
            raise ValueError('x and y cannot be both different than 0.')

        self._add(r.translate(*self.ref))

        if update_ref:
            self.ref = [self.ref[0]+x, self.ref[1]+y]

        # update bounding polygon
        a,b = r.get_bounding_box()
        bp = Rectangle((a[0], a[1]),
                        (b[0], b[1]))
        self._bounding_polygon+=bp

        return self


    def add_circular_end(self, orientation:str,
                               update_ref: bool=False,
                               nb_points: int=50) -> PolygonSet:
        """
        Add a circular open end to a coplanar waveguide in the given
        orientation.
        Notice that this method does not increment the total length of the
        waveguide nor its reference point (unless update_ref is set to True)!

        Parameters
        ----------
        orientation : str {'left', right', 'top', 'bottom'}
            Orientation of the open end.
            Also the orientation can be writen in various way, for instance
            following are equivalent:
            'l', 'left'
        update_ref : bool
            If True update the reference point of the cpw by including the length
            of the end in x and y direction.
            Useful when starting a cpw by an end_point.
        nb_point : int (default=50)
            Number of point used in the polygon.
        """

        theta = np.linspace(-np.pi/2, np.pi/2, nb_points)
        x = np.concatenate(((self._w/2.)*np.cos(theta), (self._w/2.+self._s)*np.cos(theta[::-1])))
        y = np.concatenate(((self._w/2.)*np.sin(theta), (self._w/2.+self._s)*np.sin(theta[::-1])))
        p = PolygonSet(polygons=[np.vstack((x, y)).T],
                       layers=[self._layer],
                       datatypes=[self._datatype],
                       names=[self._name],
                       colors=[self._color])

        # generate bounding polygon
        x = np.array(((self._w/2.+self._s)*np.cos(theta[::-1])))
        y = np.array(((self._w/2.+self._s)*np.sin(theta[::-1])))
        bp = PolygonSet(polygons=[np.vstack((x, y)).T])

        if orientation in ('r', 'right'):
            added_ref = (self._w/2 + self._s, 0.)

        elif orientation in ('b', 'bottom'):
            p.rotate(-np.pi/2.)
            bp.rotate(-np.pi/2.)
            added_ref = (0., -self._w/2 - self._s)

        elif orientation in ('l', 'left'):
            p.rotate(np.pi)
            bp.rotate(np.pi)
            added_ref = (-self._w/2 - self._s, 0.)

        elif orientation in ('t', 'top'):
            p.rotate(np.pi/2.)
            bp.rotate(np.pi/2.)
            added_ref = (0., -self._w/2 - self._s)

        else:
            raise ValueError('Your orientation "'+orientation+'" is incorrect.\nMust be "right", "bottom", "left" or "top".')

        self._add(p.translate(*self.ref))
        self._bounding_polygon+=bp.translate(*self.ref)

        if update_ref:
            self.ref = [self.ref[0]+added_ref[0], self.ref[1]+added_ref[1]]

        return self


    def add_fresnel_end(self, orientation: str,
                              update_ref: bool=False,
                              nb_points: int=101) -> PolygonSet:
        """
        Add a fresnel open end to a coplanar waveguide in the given
        orientation.
        Notice that this method does not increment the total length of the
        waveguide nor its reference point (unless update_ref is set to True)!

        Parameters
        ----------
        orientation : str {'left', right', 'top', 'bottom'}
            Orientation of the open end.
            Also the orientation can be writen in various way, for instance
            following are equivalent:
            'l', 'left'
        update_ref : bool
            If True update the reference point of the cpw by including the length
            of the end in x and y direction.
            Useful when starting a cpw by an end_point.
        nb_point : int (default=101)
            Number of point used in the polygon.
            Must be odd.
        """

        # Get a fresnel curve in the middle of the gap
        t = np.linspace(0, self._get_fresnel_parametric_length(np.pi), nb_points)
        x, y = self._get_fresnel_curve(np.pi, self._w+self._s, nb_points)
        dx, dy = np.gradient(x, t), np.gradient(y, t)

        # Get the external curve
        dx1, dy1 = dx, dy
        n = np.sqrt(dx1**2 + dy1**2.)
        dx1, dy1 = dx1/n, dy1/n
        x1, y1 = x, y
        theta1 = np.angle(dx1+1j*dy1)-np.pi/2.
        x1, y1 = x1+np.cos(theta1)*self._s/2., y1+np.sin(theta1)*self._s/2.

        # Get the internal curve
        dx2, dy2 = dx[::-1], dy[::-1]
        n = np.sqrt(dx2**2 + dy2**2.)
        dx2, dy2 = dx2/n, dy2/n
        x2, y2 = x[::-1], y[::-1]
        theta2 = np.angle(dx2+1j*dy2)+np.pi/2.
        x2, y2 = x2+np.cos(theta2)*self._s/2., y2+np.sin(theta2)*self._s/2.

        # Combine both curve to get a polygon and put its lower left at (0, 0)
        x, y = np.concatenate((x1, x2)), np.concatenate((y1, y2))
        p = PolygonSet(polygons=[np.vstack((x, y)).T],
                       layers=[self._layer],
                       datatypes=[self._datatype],
                       names=[self._name],
                       colors=[self._color])
        bp = PolygonSet(polygons=[np.vstack((x2, y2)).T])
        added_length = p.get_size()[0]
        bp.translate(-p.polygons[0][nb_points,0], -p.polygons[0][nb_points,1])
        p.translate(-p.polygons[0][nb_points,0], -p.polygons[0][nb_points,1])

        # Correct points to match perfectly to the width and gap
        p.polygons[0][0] = [0, self._s+self._w]
        p.polygons[0][nb_points-1] = [0, self._s]
        p.polygons[0][nb_points] = [0, 0]
        p.polygons[0][-1] = [0, 2*self._s+self._w]
        p.translate(*self.ref)
        bp.polygons[0][-1] = [0, (2*self._s+self._w)]
        bp.polygons[0][0] = [0, 0]
        bp.translate(*self.ref)

        if orientation in ('r', 'right'):
            added_ref = (added_length, 0)
            self._add(p.translate(0, -self._w/2-self._s))
            self._bounding_polygon+=bp.translate(0, -self._w/2-self._s)

        elif orientation in ('b', 'bottom'):
            p = p.rotate(-np.pi/2, center=p.get_center())
            bp = bp.rotate(-np.pi/2, center=p.get_center())
            dx, dy = (-p.polygons[0][-1,0], -p.polygons[0][-1,1])
            p.translate(dx, dy)
            p.translate(self.ref[0], self.ref[1])
            bp.translate(dx, dy)
            bp.translate(self.ref[0], self.ref[1])

            added_ref = (0, -added_length)

            self._add(p.translate(+self._w/2+self._s, 0))
            self._bounding_polygon+=bp.translate(+self._w/2+self._s, 0)

        elif orientation in ('l', 'left'):
            p = p.flip('y')
            dx, dy = (-p.polygons[0][nb_points,0], -p.polygons[0][nb_points,1])
            p.translate(dx, dy)
            p.translate(self.ref[0], self.ref[1])
            bp.flip('y')
            bp.translate(dx, dy)
            bp.translate(self.ref[0], self.ref[1])

            added_ref = (-added_length, 0)

            self._add(p.translate(0, -self._w/2-self._s))
            self._bounding_polygon+=bp.translate(0, -self._w/2-self._s)

        elif orientation in ('t', 'top'):
            p = p.rotate(np.pi/2, center=p.get_center())
            bp = bp.rotate(np.pi/2, center=p.get_center())
            dx, dy = (-p.polygons[0][-1,0], -p.polygons[0][-1,1])
            p.translate(dx, dy)
            bp.translate(dx, dy)
            p.translate(self.ref[0], self.ref[1])
            bp.translate(self.ref[0], self.ref[1])

            added_ref = (0, added_length)

            self._add(p.translate(-self._w/2-self._s, 0))
            self._bounding_polygon+=bp.translate(-self._w/2-self._s, 0)

        else:
            raise ValueError('Your orientation "'+orientation+'" is incorrect.\nMust be "right", "bottom", "left" or "top".')

        if update_ref:
            self.ref = [self.ref[0]+added_ref[0], self.ref[1]+added_ref[1]]

        return self



from functools import lru_cache, partial
import numpy as np
from typing import Optional, Union, Tuple, Self
from scipy.special import fresnel
from typing import Literal, Dict, Callable
from scipy.interpolate import InterpolatedUnivariateSpline
from scipy.integrate import quad
from scipy.special import fresnel
from scipy.optimize import fsolve
_mpone = np.array((-1.0, 1.0))


from pygdsdesign.polygonSet import PolygonSet
from pygdsdesign.typing_local import Coordinate



class TransmissionLine(PolygonSet):


    def __init__(
        self,
        layer: int,
        datatype: int,
        name: str,
        color: str,
        ref: Optional[Coordinate] = None,
    ) -> None:
        """
        TransmissionLine class for strip classes allowing easy draw of complex circuits.

        Parameters
        ----------
        layer : int
            Layer number of the coplanar.
        datatype : int
            Datatype number of the coplanar.
        ref : List[float, float] (default None)
            Coordinate of the last points.
            If None, corresponds to [0, 0].
        """

        # Avoid mutability of the default reference by instancing the list inside the object
        if ref is None:
            self.ref: Coordinate = [0., 0.]
        else:
            self.ref = ref

        PolygonSet.__init__(self, polygons=[np.array([self.ref])],
                                  layers=[layer],
                                  datatypes=[datatype],
                                  names=[name],
                                  colors=[color])
        self._layer = layer
        self._datatype = datatype
        self._name = name
        self._color = color

        self.total_length = 0. # Total length of the strip

        # Store computed Fresnel information to save computation timne
        computedSerpentine: Dict[Tuple[float, int, float],
                                 Tuple[np.ndarray, np.ndarray,
                                       np.ndarray,
                                       np.ndarray, np.ndarray,
                                       Callable, Callable]] = {}


    @property
    def layer(self):
        self._layer


    @layer.setter
    def layer(self, layer:int):

        self.layers = [layer]*len(self.polygons)
        self._layer = layer


    @property
    def datatype(self):
        self._datatype


    @datatype.setter
    def datatype(self, datatype:int):

        self.datatypes = [datatype]*len(self.polygons)
        self._datatype = datatype


    @property
    def name(self):
        self._name


    @name.setter
    def name(self, name:str):

        self.names = [name]*len(self.polygons)
        self._name = name


    @property
    def color(self):
        self._color


    @color.setter
    def color(self, color:str):

        self.colors = [color]*len(self.polygons)
        self._color = color


    def _rot(self, x: float,
                   y: float,
                   theta: float) -> Coordinate:
        """
        Rotate set of point by theta in respect to (0, 0).
        """

        tx =  np.cos(theta)*x + np.sin(theta)*y
        ty = -np.sin(theta)*x + np.cos(theta)*y

        return tx, ty


    @partial
    def func(t, args, x, y):
        if t[0]>t[-1]:
            return x, y
        else:
            return x[::-1], y[::-1]


    @partial
    def dfunc(t, args, dx, dy,f_dx, f_dy):
        if isinstance(t, float):
            return f_dx(t), f_dy(t)
        else:
            if t[0]>t[-1]:
                return dx, dy
            else:
                return dx[::-1], dy[::-1]


    @lru_cache
    def calc(self, radius: float,
                   nb_points: int) -> Tuple[np.ndarray, np.ndarray,
                                            np.ndarray,
                                            np.ndarray, np.ndarray,
                                            Callable, Callable,
                                            float]:

        # Calculate curve length
        def curve_length(t):
            return np.hypot(f_dx(t), f_dy(t))

        t = np.linspace(0, self._get_fresnel_parametric_length(np.pi/2.), nb_points)
        x, y = self._get_fresnel_curve(np.pi/2, radius, nb_points)
        dx, dy = np.gradient(x, t), np.gradient(y, t)

        # We ensure flat derivative to get straight structure
        dx[0], dy[-1] = 0, 0

        f_dx = InterpolatedUnivariateSpline(t, dx)
        f_dy = InterpolatedUnivariateSpline(t, dy)

        length = quad(curve_length, t[0], t[-1])[0]

        return dx, dy, t, x, y, f_dx, f_dy, length



    def translate(self, dx: float,
                        dy: float) -> PolygonSet:
        """
        Translate the polygons by the amount dx, dy in the x and y direction.
        Take care of updating the reference point of the lines.
        Take care of updating the parametric curve of the lines.

        Args:
            dx (float): amount of translation in the x direction in um.
            dy (float): amount of translation in the y direction in um.

        Returns:
            PolygonSet: translated polygons.
        """

        self.ref[0] += dx
        self.ref[1] += dy

        if hasattr(self, '_param_curve'):
            self._param_curve[:,0] += dx
            self._param_curve[:,1] += dy

        if hasattr(self, '_bounding_polygon'):
            vec = np.array((dx, dy))
            self._bounding_polygon.polygons = [points + vec for points in self._bounding_polygon.polygons]

        return super().translate(dx, dy)


    def rotate(self, angle: float,
                     center: Coordinate=(0, 0)) -> Self:
        """
        Rotate this object.

        Args:
        angle : number
            The angle of rotation (in *radians*).
        center : array-like[2]
            Center point for the rotation.

        Returns:
            PolygonSet: the rotated transmission line
        """
        ca = np.cos(angle)
        sa = np.sin(angle) * _mpone
        c0 = np.array(center)
        new_polys = []

        for points in self.polygons:
            pts = points - c0
            new_polys.append(pts * ca + pts[:, ::-1] * sa + c0)
        self.polygons = new_polys

        x = np.copy(self.ref[0])-center[0]
        y = np.copy(self.ref[1])-center[1]

        self.ref[0] = np.cos(angle)*x - np.sin(angle)*y + center[0]
        self.ref[1] = np.cos(angle)*y + np.sin(angle)*x + center[1]

        return self


    ###########################################################################
    #
    #                   Add polygons to the existing polygons
    #
    ###########################################################################


    def _add(self, r: PolygonSet) -> None:
        """
        Merge polygons to existing microstrip line.
        """

        self += r


    ###########################################################################
    #
    #                   Add points to internal parametric curve
    #
    ###########################################################################


    def _add2param(self, x: Union[np.ndarray, list],
                         y: Union[np.ndarray, list],
                         t: Union[np.ndarray, list]) -> None:
        """
        Add points to internal parametric curve

        Args:
            x: x coordinate in um
            y: y coordinate in um
            t: distance along the parametric curve in um
        """



        if not hasattr(self, '_param_curve'):
            self._param_curve = np.vstack((x, y, t)).T
        else:
            new_values = np.array([x, y, self._param_curve[:,2][-1]+t]).T

            self._param_curve = np.concatenate((self._param_curve, new_values), axis=0)


    ###########################################################################
    #
    #                         Fresnel turn
    #
    ###########################################################################


    def _get_fresnel_parametric_length(self, angle: float) -> float:
        """
        Return the parametric length of the curve corresponding to half the
        given angle.

        Parameters
        ----------
        angle : float
            Angle in radian
        """

        # Pre-computer value for standard angle
        # This angle is used a lot to compute fresnel turn
        if angle==np.pi/2:
            return 0.710657090720168
        else:
            def func(t0: float,
                    target_angle: float) -> float:

                # Get the x coordinate corresponding to some parametric length
                y1, x1 = fresnel(t0)
                y0, x0 = fresnel(t0*0.99)

                return target_angle - np.angle((x1-x0) + 1j*(y1-y0))

            # Get the first half of the Fresnel curve going to 0deg to half of the
            # desired angle
            return fsolve(func, 0.75, args=np.abs(angle)/2.)[0]


    def _get_fresnel_curve(self, angle: float,
                                 radius: float,
                                 nb_points: int) -> Tuple[np.ndarray, np.ndarray]:
        """
        Return the parametric Fresnel curve along the length t.
        The curve is doubled, mirrored, joint and normalized to a given radius

        Parameters
        ----------
        angle : float
            Total angle of the curve in radian
            Must be pi or pi/2
        radius : float
            Radius of the curve
        nb_points : int
                Number of point in the polygon
        """

        if angle!=np.pi and angle!=np.pi/2:
            raise ValueError('Fresnel curve are only availabe for 90 and 180 degrees turn.')
        if nb_points%2==0:
            raise ValueError('nb_points must be an odd number.')

        t0 = self._get_fresnel_parametric_length(angle)
        t = np.linspace(0, t0, int(nb_points/2)+1)

        y1, x1 = fresnel(t)

        # Get the second half that we have to move and snap to the first one
        # Center
        x2 = x1 - x1.max()
        y2 = y1 - y1.max()

        # Mirror symetry on y
        x2 = +x2
        y2 = -y2

        # Rotation
        if angle==np.pi/2:
            x2a, y2a = self._rot(x2, y2, np.pi/2)
        else:
            x2a, y2a = x2, y2

        # Translation
        x2a += x1[-1]
        y2a += y1[-1]

        # Assembling
        y = np.concatenate((y1[:-1], y2a))
        x = np.concatenate((x1[:-1], x2a))

        # Sorting along the y direction
        y_t = np.argsort(y)
        y = y[y_t]
        x = x[y_t]

        # Normalize to the correct radius

        # Get the longest direction
        ym = abs(y.max()-y.min())
        xm = abs(x.max()-x.min())

        if xm>=ym:
            r = xm
        else:
            r = ym

        y = y/r*radius+self._w/2.
        x = x/r*radius

        return x[::-1], y[::-1]


    ###########################################################################
    #
    #                   Meta method to create complex microstrip path
    #
    ###########################################################################


    def add_serpentine(self, total_length: float,
                             nb_turn: int,
                             spacing: float,
                             orientation: Literal['right', 'left', 'top', 'bottom']='right',
                             starting: Literal['left', 'right']='left',
                             spacing_mode: Literal['center2center', 'edge2edge']='center2center',
                             turn: Literal['circular', 'fresnel']='circular',
                             nb_points_per_turn: Optional[int]=None) -> PolygonSet:
        """
        Add a microstrip serpentine.
        The spacing between the strip can be calculated from strip-center to
        strip-center or strip-edge to strip-edge through the spacing_mode
        parameter

        Parameters
        ----------
        total_length : float
            Total length of the serpentine in um.
        nb_turn : int
            Number of serpentine turn.
        spacing : float
            Spacing between the strip in um.
            This spacing is calculated through two different modes, see
            spacing_mode.
        orientation : str {'right', 'left', 'top', 'bottom'} (default 'right')
            Orientation of the serpentine.
        starting : str {'right', 'left'} (default 'left')
            Orientation of the first turn of the serpentine in respect to the
            last microstrip segment.
        spacing_mode : str {'center2center', 'edge2edge'}, (default 'center2center')
            The spacing between the strip can be calculated from strip-center to
            strip-center or strip-edge to strip-edge depending of this parameter.
        turn : str {'circular', 'fresnel'}, (default 'circular')
            Type or turn.
        nb_points_per_turn : Optional(int) (default None)
            Number of point used per turn.
            If None, used the default number of point of the circular, 50, or
            Fresnel, 51, turn.
        """

        # Calculate the radius depending of the spacing mode
        if spacing_mode == 'center2center':
            r = spacing/2.
        elif spacing_mode == 'edge2edge':
            r = spacing/2. + self._w/2.
        else:
            raise ValueError('"spacing_mode" must be in ["center2center", "edge2edge"]')

        # Depending of the turn mode, the turn function and the turn length vary
        if turn not in ('circular', 'fresnel'):
            raise ValueError('"turn" must be in ["circular", "fresnel"]')
        else:
            if turn=='circular':
                if nb_points_per_turn is None:
                    nb_points_per_turn = 50

                turn_length =  np.pi*r/2
                turn_func = self.add_turn
            elif turn=='fresnel':
                if nb_points_per_turn is None:
                    nb_points_per_turn = 51

                # All the computations are stored in self.calc to cache the result
                dx, dy, t, x, y, f_dx, f_dy, turn_length = self.calc(r, 1001)

                turn_func = self.add_fresnel_turn

        # Calculate the serpentine length
        s_length = (total_length-(2*nb_turn+2)*turn_length+2*r)/2/nb_turn

        # Prepare variable to build serpentine depending on its orientation
        if orientation.lower() in ['right', 'left']:
            c = 'b'
            d = 't'
            x1 = 0.
            x2 = 0.
            y1 = s_length-r
            y2 = 2.*s_length

            if orientation.lower() == 'right':
                a = 'l'
                b = 'r'
            elif orientation.lower() == 'left':
                a = 'r'
                b = 'l'
                if starting.lower()=='left':
                    starting = 'right'
                else:
                    starting = 'left'
        elif orientation.lower() in ['top', 'bottom']:
            c = 'l'
            d = 'r'
            x1 = s_length-r
            x2 = 2.*s_length
            y1 = 0.
            y2 = 0.
            if orientation.lower()=='top':
                a = 'b'
                b = 't'
                if starting.lower()=='left':
                    starting = 'right'
                else:
                    starting = 'left'
            elif orientation.lower()=='bottom':
                a = 't'
                b = 'b'
        else:
            raise ValueError('"orientation" must be "right", "left", "top", "right".')

        # Start
        if starting.lower()=='left':
            turn_func(r, a+d, nb_points=nb_points_per_turn)
            self.add_line(x1, y1)
            counter = 0
            final = nb_turn - 1
        elif starting.lower()=='right':
            turn_func(r, a+c, nb_points=nb_points_per_turn)
            self.add_line(-x1, -y1)
            counter = 1
            final = nb_turn
        else:
            raise ValueError('starting must be "left" or "right".')

        # Middle
        while counter<final:
            if counter%2:
                turn_func(r, d+b, nb_points=nb_points_per_turn)
                turn_func(r, a+d, nb_points=nb_points_per_turn)
                self.add_line(x2, y2)
            else:
                turn_func(r, c+b, nb_points=nb_points_per_turn)
                turn_func(r, a+c, nb_points=nb_points_per_turn)
                self.add_line(-x2, -y2)

            counter += 1

        # End from either to or bottom
        if counter%2:
            turn_func(r, d+b, nb_points=nb_points_per_turn)
            turn_func(r, a+d, nb_points=nb_points_per_turn)
            self.add_line(x1, y1)
            turn_func(r, c+b, nb_points=nb_points_per_turn)
        else:
            turn_func(r, c+b, nb_points=nb_points_per_turn)
            turn_func(r, a+c, nb_points=nb_points_per_turn)
            self.add_line(-x1, -y1)
            turn_func(r, d+b, nb_points=nb_points_per_turn)

        return self




from pygdsdesign import (
    GdsLibrary,
    PolygonSet,
    Rectangle,
)
from pygdsdesign.shapes import lateqs_logo
from pygdsdesign.operation import (
    offset,
    grid_cover,
    subtraction,
)

from copy import deepcopy

# define layer
layer_background = {
    "layer": 1,
    "name": "background",
    "datatype": 0,
    "color": "#ff00ff",
}

layer_grid = {
    "layer": 2,
    "name": "grid",
    "datatype": 0,
    "color": "#13c24b",
}


# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell("TOP")

# Create empty polygon to which other polygon will be added
tot = PolygonSet()

# the polygon we will operate on
poly = lateqs_logo()

# Make a rectangle big enough for our purposes
r = offset(
    Rectangle(*lateqs_logo().get_bounding_box()),
    20,
)

# make the negative of the group logo
r1 = subtraction(r, poly, **layer_background)
r2 = deepcopy(r1).translate(500, 500)

# Use that negative for the grid cover operation
grid1 = grid_cover(
    polygons=r1,
    square_width=12,
    square_gap=23,
    safety_margin=10,
    centered=True,
    noise=12,
    only_square=True,
    **layer_grid
)


# Same with hexagon
grid2 = grid_cover(
    polygons=r2,
    square_width=10,
    square_gap=10,
    safety_margin=10,
    centered=True,
    hexagonal_grid=True,
    noise=3,
    only_square=True,
    **layer_grid
)



tot += r1 + r2 + grid1 + grid2

# Add polygons to cell
cell.add(tot)

lib.export_gds("gridcover.gds")


import numpy as np
from copy import copy,deepcopy

from pygdsdesign import (GdsLibrary, PolygonSet, Text, CPWPolar)

def spiral(u, args):

    r = args[1]*2.*args[2]*u + args[0]/2.
    theta = args[1]*2.*np.pi*u
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return x, y


def dspiral_dt(u, args):

    r = args[1]*2.*args[2]*u
    theta = args[1]*2.*np.pi*u
    dx_dt = -args[1]*2.*np.pi * \
        np.sin(theta)*r + np.cos(theta)*args[1]*2.*args[2]
    dy_dt = args[1]*2.*np.pi*np.cos(theta)*r + np.sin(theta)*args[1]*2.*args[2]
    return dx_dt, dy_dt


def circle(t, args):
    x = args[0] * np.cos(t)
    y = args[0] * np.sin(t)
    return x, y

def dcircle_dt(t, args):
    dx_dt = - args[0] * np.sin(t)
    dy_dt = + args[0] * np.cos(t)
    return dx_dt, dy_dt


# Chip size
chip_dx = 595*2
chip_dy = 767*2

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('TOP')


# Create empty polygon to which other polygon will be added
tot = PolygonSet()


# Add label on the top right of the chip
t = Text('CPW_Polar', 250)
dx_t, dy_t = t.get_size()
tot += t.translate(-dx_t+chip_dx, -dy_t-250)



# Center the design to the chip and the chip to the (0,0) of the cell

largeur_micro=1
m = CPWPolar(width=largeur_micro, gap=largeur_micro/5, angle=-np.pi/2.24654)

m.add_line(3)

u = np.linspace(0, 1, 10000)
inner_diameter = 10
nb_turn = 4
spiral_spacing = 3.
spiral_width = 2.
m.add_parametric_curve(spiral, dspiral_dt, u, args=(inner_diameter, nb_turn, spiral_spacing))

m.add_line(1)
m.add_line(3)
m.add_line(1)
m.add_line(3)
m.add_turn(3, np.pi/3)
m.add_line(3)
m.add_turn(3, -np.pi/2.56)
m.add_line(3)
m.add_taper(3, largeur_micro*2, largeur_micro/2)
m.add_turn(3, -np.pi/2.56)
m.add_line(3)
m.add_turn(3, np.pi/2.56)
m.add_taper(3, largeur_micro, largeur_micro/5)
m.add_turn(3, -np.pi/2.56)
m.add_line(3)
m.add_parametric_curve(circle, dcircle_dt,np.linspace(-np.pi/2.24, np.pi/0.55412 , 100), args=[0.732854],add_polygon=True)
m.add_line(3)
m.add_turn(3, -np.pi/0.512)
m.add_turn(3, +np.pi/1.2124)
m.add_line(3)
m.add_parametric_curve(circle, dcircle_dt, np.linspace(-np.pi /2.24, np.pi/1.56451, 100), args=[1.15], add_polygon=True)
m.add_line(3)
m.add_parametric_curve(circle, dcircle_dt,np.linspace(-np.pi/2.24, np.pi/0.55412 , 100), args=[0.732854],add_polygon=True)
m.add_line(3)
m.add_parametric_curve(circle, dcircle_dt,np.linspace(-np.pi/3.65, np.pi/2.55412 , 100), args=[1.4854],add_polygon=True)
m.add_taper(3, largeur_micro/2, largeur_micro/10)
m.add_end(0.02, True)
m.add_line(3)
m.add_turn(3, +np.pi/2.2124)
m.add_circular_end(nb_points=101, update_ref=True)
m.add_line(3)

#adding bounding polygon
b_p = CPWPolar(width=largeur_micro, gap=largeur_micro/5, angle=-np.pi/2.24654,layer=42)
b_p+=m.bounding_polygon.change_layer(42)
cell.add(b_p)
tot+=m

tot+=m




cell.add(tot)
# Save GDS file
lib.write_gds('cpw_polar.gds')


from typing import Union
from pygdsdesign.polygonSet import PolygonSet
from pygdsdesign.operation import select_polygon_per_layer
import numpy as np

def print_ebeam_time(polygon: PolygonSet,
                     layer: Union[int, dict],
                     beam_current: float,
                     resist_dose: float,
                     datatype: int=0,
                     eos: int=3,
                     stage_displacement_time: float=0.05) -> None:
    """
    Print the exposure time of the layer based on its area, the beam current and
    the resist dose.
    Print the stage displacement time of the layer based on its area, the eos
    mode and the estimated stage displacement time.
    The total time is simply the sum of the exposure and displacement time.

    Args:
        polygon: Polygon from which the layer is extracted
        layer:
            if int: layer we want the ebeam exposure time from
            if dict if a key 'layer': layer we want the ebeam exposure time from
        beam_current: beam current in nA.
        resist_dose: resis dose in uC/cm2
        datatype: layer datatype.
            if argument `layer` is a dict, this argument `datatype` is ignored.
            if argument `layer` is an int, this argument `datatype` is used.
            default, 0.
        eos: EOS mode of the ebeam job.
            Must be either 3 or 6.
            An eos mode 3 imply a stage field of 500x500 um2.
            An eos mode 6 imply a stage field of 62.5x62.5 um2.
        stage_displacement_time: Time taken by the ebeam stage to move from
            field to field in second.
    """
    if isinstance(layer, dict):
        if 'layer' in layer.keys():
            if isinstance(layer['layer'], int):
                l = layer['layer']
            else:
                raise ValueError('The key "layer" of the layer argument must be an int')
        else:
            raise ValueError('Layer argument must be a dict containing a key "layer"')
        if 'datatype' in layer.keys():
            if isinstance(layer['datatype'], int):
                d = layer['datatype']
            else:
                raise ValueError('The key "datatype" of the layer argument must be an int')
        else:
            raise ValueError('Layer argument must be a dict containing a key "datatype"')
    elif isinstance(layer, int):
        l = layer
        d = datatype
    else:
        raise ValueError('Layer argument must be an int')

    # get layer area in um2
    area = select_polygon_per_layer(polygon, layer=l, datatype=d).get_area()

    # exposure time in s
    exposure_time = area/1e8*resist_dose*1e3/beam_current
    minutes = exposure_time//60
    hours = minutes//60

    print('')
    print('++++++++++++++++++++++++++++++++++++')
    print('ebeam info:')
    print('    resist sensitivity: {} uC/cm2'.format(resist_dose))
    print('    area: {:.0f} um2 = {:.1f} mm2'.format(area, area/1e6))
    print('    current: {} nA'.format(beam_current))
    print('    exposure duration: {:2.0f}h {:2.0f}min {:2.0f}s'.format(hours, minutes % 60, exposure_time % 60))

    # get layer containing area in um2
    dx, dy = polygon.get_size()
    area_c = dx*dy
    # stage displacement time
    if eos==3:
        field_area = 500.*500.
    elif eos==6:
        field_area = 62.5*62.5
    else:
        raise ValueError('eos argument must be either 3 or 6')

    stage_time = area_c/field_area*stage_displacement_time
    minutes = stage_time//60
    hours = minutes//60
    print('    number of field: {:.0f}'.format(area_c/field_area))
    print('    stage displacement duration: {:2.0f}h {:2.0f}min {:2.0f}s'.format(hours, minutes % 60, stage_time % 60))

    # total time
    total_time = exposure_time + stage_time
    minutes = total_time//60
    hours = minutes//60
    print('    total ebeam duration: {:2.0f}h {:2.0f}min {:2.0f}s'.format(hours, minutes % 60, total_time % 60))




def distance(x1:float,y1:float,x2:float,y2:float) -> float:
            """Return the distance between the two point (x1;y1) and (x2;y2)"""
            return np.sqrt((x1-x2)**2 + (y2-y1)**2)


from math import ceil
from typing import Any, Dict, List, Literal, Tuple, Optional
import warnings
import copy as libcopy
import numpy as np

import clipper
from pygdsdesign.polygons import Rectangle
from pygdsdesign.polygonSet import PolygonSet
from pygdsdesign.typing_local import Coordinates


def boolean(operand1: PolygonSet,
            operand2: PolygonSet|list,
            operation: Literal['or', 'and', 'xor', 'not'],
            precision: float=0.001,
            layer: int=0,
            datatype: int=0,
            name: str ='',
            color: str ='') -> Optional[PolygonSet]:
    """
    Execute any boolean operation between 2 polygons or polygon sets.

    Parameters
    ----------
    operand1 : `PolygonSet`, `CellReference`, `CellArray` or iterable
        First operand.  If this is an iterable, each element must be a
        `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.
    operand2 : None, `PolygonSet`, `CellReference`, `CellArray` or iterable
        Second operand.  If this is an iterable, each element must be a
        `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.
    operation : {'or', 'and', 'xor', 'not'}
        Boolean operation to be executed.  The 'not' operation returns
        the difference ``operand1 - operand2``.
    precision : float
        Desired precision for rounding vertice coordinates.
    layer : integer
        The GDSII layer number for the resulting element.
    datatype : integer
        The GDSII datatype for the resulting element (between 0 and
        255).

    Returns
    -------
    out : PolygonSet or None
        Result of the boolean operation.
    """
    poly1 = _gather_polys(operand1)
    poly2 = _gather_polys(operand2)

    if len(poly2) == 0:
        if operation == "not" or operation=="xor":
            if len(poly1) == 0:
                warnings.warn("[pygdsdesign] You try to 'boolean' on a empty polygonSet",stacklevel=4,)
                return None
            return PolygonSet(poly1, operand1.layers, operand1.datatypes, operand1.names, operand1.colors)
        poly2.append(poly1.pop())

    result = clipper.clip(poly1, poly2, operation, 1 / precision)

    if len(result) == 0:
        warnings.warn("[pygdsdesign] The result of your boolean operation is None",stacklevel=4,)
        return None

    return PolygonSet(result, [layer]*len(result), [datatype]*len(result), [name]*len(result), [color]*len(result))


def addition(operand1: PolygonSet,
             operand2: PolygonSet,
             precision: float=0.001,
             layer: int=0,
             datatype: int=0,
             name: str ='',
             color: str ='') -> Optional[PolygonSet]:
    """
    Execute the OR boolean operation between 2 polygon sets.

    Parameters
    ----------
    operand1 : `PolygonSet`, `CellReference`, `CellArray` or iterable
        First operand.  If this is an iterable, each element must be a
        `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.
    operand2 : None, `PolygonSet`, `CellReference`, `CellArray` or iterable
        Second operand.  If this is an iterable, each element must be a
        `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.
    precision : float
        Desired precision for rounding vertice coordinates.
    layer : integer
        The GDSII layer number for the resulting element.
    datatype : integer
        The GDSII datatype for the resulting element (between 0 and
        255).

    Returns
    -------
    out : PolygonSet or None
        Result of the boolean operation.
    """
    return boolean(operand1, operand2, "or", precision, layer, datatype, name, color)


def subtraction(operand1: PolygonSet,
                 operand2: PolygonSet,
                 precision: float=0.001,
                 layer: int=0,
                 datatype: int=0,
                 name: str ='',
                 color: str ='') -> Optional[PolygonSet]:
    """
    Execute the NOT boolean operation between 2 polygon sets.

    Parameters
    ----------
    operand1 : `PolygonSet`, `CellReference`, `CellArray` or iterable
        First operand.  If this is an iterable, each element must be a
        `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.
    operand2 : None, `PolygonSet`, `CellReference`, `CellArray` or iterable
        Second operand.  If this is an iterable, each element must be a
        `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.
    precision : float
        Desired precision for rounding vertice coordinates.
    layer : integer
        The GDSII layer number for the resulting element.
    datatype : integer
        The GDSII datatype for the resulting element (between 0 and
        255).

    Returns
    -------
    out : PolygonSet or None
        Result of the boolean operation.
    """
    return boolean(operand1, operand2, "not", precision, layer, datatype, name, color)


def intersection(operand1: PolygonSet,
                 operand2: PolygonSet,
                 precision: float=0.001,
                 layer: int=0,
                 datatype: int=0,
                 name: str ='',
                 color: str ='') -> Optional[PolygonSet]:
    """
    Execute the AND boolean operation between 2 polygon sets.

    Parameters
    ----------
    operand1 : `PolygonSet`, `CellReference`, `CellArray` or iterable
        First operand.  If this is an iterable, each element must be a
        `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.
    operand2 : None, `PolygonSet`, `CellReference`, `CellArray` or iterable
        Second operand.  If this is an iterable, each element must be a
        `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.
    precision : float
        Desired precision for rounding vertice coordinates.
    layer : integer
        The GDSII layer number for the resulting element.
    datatype : integer
        The GDSII datatype for the resulting element (between 0 and
        255).

    Returns
    -------
    out : PolygonSet or None
        Result of the boolean operation.
    """
    return boolean(operand1, operand2, "and", precision, layer, datatype, name, color)


def difference(operand1: PolygonSet,
               operand2: PolygonSet,
               precision: float=0.001,
               layer: int=0,
               datatype: int=0,
               name: str ='',
               color: str ='') -> Optional[PolygonSet]:
    """
    Execute the XOR boolean operation between 2 polygon sets.

    Parameters
    ----------
    operand1 : `PolygonSet`, `CellReference`, `CellArray` or iterable
        First operand.  If this is an iterable, each element must be a
        `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.
    operand2 : None, `PolygonSet`, `CellReference`, `CellArray` or iterable
        Second operand.  If this is an iterable, each element must be a
        `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.
    precision : float
        Desired precision for rounding vertice coordinates.
    layer : integer
        The GDSII layer number for the resulting element.
    datatype : integer
        The GDSII datatype for the resulting element (between 0 and
        255).

    Returns
    -------
    out : PolygonSet or None
        Result of the boolean operation.
    """
    return boolean(operand1, operand2, "xor", precision, layer, datatype, name, color)


def crop(
    polygons: PolygonSet,
    orientation: Literal["top", "right", "bottom", "left"],
    value: float,
    layer: int = 0,
    datatype: int = 0,
    name: str = "",
    color: str = "",
) -> Optional[PolygonSet]:
    """
    Crop a polygon or a collection of polygon by a certain value from a certain
    direction

    Args:
        orientation: the orientation of the crop.
            Must be 'top', 'right', 'bottom' or 'left'.
        value: value of the crop in um
    """
    bounding_box = polygons.get_bounding_box()

    if bounding_box is None:
        warnings.warn("[pygdsdesign] Can't crop a polygon with a null area.")
        return None

    ((x0, y0), (x1, y1)) = bounding_box

    if orientation=='top':
        t = Rectangle((x0, y0), (x1, y1-value))
    elif orientation=='right':
        t = Rectangle((x0, y0), (x1-value, y1))
    elif orientation=='bottom':
        t = Rectangle((x0, y0+value), (x1, y1))
    elif orientation=='left':
        t = Rectangle((x0+value, y0), (x1, y1))
    else:
        raise ValueError("orientation must be 'top', 'right', 'bottom' or 'left'")

    return boolean(polygons, t, 'and', layer=layer, datatype=datatype, name=name, color=color)


def merge(polygons) -> PolygonSet:
    """
    Merge polygons of the same layer and datatype
    """
    # Sort all the polygons per layer
    layers: Dict[Tuple[int, int, str, str], List[np.ndarray]] = {}

    for poly, layer, datatype, name, color in zip(polygons.polygons, polygons.layers, polygons.datatypes, polygons.names, polygons.colors):
        if (layer, datatype, name, color) in layers:
            layers[(layer, datatype, name, color)].append(poly)
        else:
            layers[(layer, datatype, name, color)] = [poly]

    # Merge all polygons layer per layer
    tot = PolygonSet(
        polygons=[np.array([[0.0, 0.0]])],
        layers=[layer],
        datatypes=[datatype],
        names=[name],
        colors=[color],
    )

    for (layer, datatype, name, color), poly in layers.items():
        temp = boolean(poly, [],
                       'or',
                       layer=layer,
                       datatype=datatype,
                       name=name,
                       color=color)
        if temp is not None:
            tot += temp

    return tot


def _gather_polys(args):
    """
    Gather polygons from different argument types into a list.

    Parameters
    ----------
    args : None, `PolygonSet`, `CellReference`, `CellArray` or iterable
        Polygon types.  If this is an iterable, each element must be a
        `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.

    Returns
    -------
    out : list of numpy array[N][2]
        List of polygons.
    """
    if args is None:
        warnings.warn("[pygdsdesign] You try to '_gather_polys' on empty polygonSets",stacklevel=4)
        return []

    if isinstance(args, PolygonSet):
        return [p for p in args.polygons]

    polys = []

    for p in args:
        if isinstance(p, PolygonSet):
            polys.extend(p.polygons)
        else:
            polys.append(p)

    return polys


def slice(polygons, position, axis, precision=1e-3, layer=0, datatype=0):
    """
    Slice polygons and polygon sets at given positions along an axis.

    Parameters
    ----------
    polygons : `PolygonSet`, `CellReference`, `CellArray` or iterable
        Operand of the slice operation.  If this is an iterable, each
        element must be a `PolygonSet`, `CellReference`, `CellArray`,
        or an array-like[N][2] of vertices of a polygon.
    position : number or list of numbers
        Positions to perform the slicing operation along the specified
        axis.
    axis : 0 or 1
        Axis along which the polygon will be sliced.
    precision : float
        Desired precision for rounding vertice coordinates.
    layer : integer, list
        The GDSII layer numbers for the elements between each division.
        If the number of layers in the list is less than the number of
        divided regions, the list is repeated.
    datatype : integer, list
        The GDSII datatype for the resulting element (between 0 and
        255).  If the number of datatypes in the list is less than the
        number of divided regions, the list is repeated.

    Returns
    -------
    out : list[N] of `PolygonSet` or None
        Result of the slicing operation, with N = len(positions) + 1.
        Each PolygonSet comprises all polygons between 2 adjacent
        slicing positions, in crescent order.

    Examples
    --------
    >>> ring = Round((0, 0), 10, inner_radius = 5)
    >>> result = slice(ring, [-7, 7], 0)
    >>> cell.add(result[1])
    """
    polys = _gather_polys(polygons)
    if not isinstance(layer, list):
        layer = [layer]
    if not isinstance(datatype, list):
        datatype = [datatype]
    if not isinstance(position, list):
        pos = [position]
    else:
        pos = sorted(position)

    result = [[] for _ in range(len(pos) + 1)]
    scaling = 1 / precision

    for pol in polys:
        for r, p in zip(result, clipper._chop(pol, pos, axis, scaling)):
            r.extend(p)

    for i in range(len(result)):
        if len(result[i]) == 0:
            result[i] = None
        else:
            result[i] = PolygonSet(
                result[i], layer[i % len(layer)], datatype[i % len(datatype)]
            )

    return result


def offset(polygons: PolygonSet,
           distance: float,
           join: str='miter',
           tolerance: float=2,
           precision: float=0.001,
           join_first=False,
           layer: int=0,
           datatype: int=0,
           name: str='',
           color: str='') -> Optional[PolygonSet]:
    """
    Shrink or expand a polygon set.

    Parameters
    ----------
    polygons : `PolygonSet`, `CellReference`, `CellArray` or iterable
        Polygons to be offset.  If this is an iterable, each element
        must be a `PolygonSet`, `CellReference`, `CellArray`, or an
        array-like[N][2] of vertices of a polygon.
    distance : number
        Offset distance.  Positive to expand, negative to shrink.
    join : 'miter', 'bevel', 'round'
        Type of join used to create the offset polygon.
    tolerance : number
        For miter joints, this number must be at least 2 and it
        represents the maximal distance in multiples of offset between
        new vertices and their original position before beveling to
        avoid spikes at acute joints.  For round joints, it indicates
        the curvature resolution in number of points per full circle.
    precision : float
        Desired precision for rounding vertex coordinates.
    join_first : bool
        Join all paths before offsetting to avoid unnecessary joins in
        adjacent polygon sides.
    layer : integer
        The GDSII layer number for the resulting element.
    datatype : integer
        The GDSII datatype for the resulting element (between 0 and
        255).

    Returns
    -------
    out : `PolygonSet` or None
        Return the offset shape as a set of polygons.
    """
    result = clipper.offset(
        _gather_polys(polygons),
        distance,
        join,
        tolerance,
        1 / precision,
        1 if join_first else 0,
    )

    if len(result) == 0:
        warnings.warn("[pygdsdesign] The result of 'offset' is None",stacklevel=4,)
        return None

    return PolygonSet(result, [layer]*len(result), [datatype]*len(result), [name]*len(result), [color]*len(result))


def grid_cover(polygons: PolygonSet,
               square_width: float=10,
               square_gap: float=12,
               safety_margin: float=10,
               centered: bool=False,
               noise: float=0,
               only_square: bool=True,
               hexagonal_grid: bool=False,
               layer: int=1,
               datatype: int=0,
               name: str='',
               color: str ='') ->  PolygonSet:
    """
        Create a grid pattern of squares which follows any shape given in polygons
        after shrinking it by a safety margin.
        The square size and gap between them are given as free parameters.

        If hexagonal_grid is True, use hexagon instead. see hexagonal_grid below.

        Args:
            polygons: shape from which the grid pattern is built upon.
            square_width: Width of the square in um.
                Defaults to 10um.
            square_gap: Space between the square in um.
                Defaults to 12um.
            safety_margin: shrinking scaling distance taken from the polygon in um.
                Defaults to 10um.
            centered: if False, the grid bottom left corresponds to the polygons
                bottom left.
                If True, the grid center corresponds to the polygon center.
                Defaults to False.
            noise: Noise parameter in um to generate random square positions.
                Each square is translated by a random amount given by the noise
                parameter. A noise value uf 4 um will give a random shift
                from -4 um to +4 um in x and y.
                Defaults to 0, resulting in a regular grid.
            only_square: Remove all polygons that are not stricly
                `square_width` x `square_width` shape.
                Defaults to True.
            hexagonal_grid: Replace the square grid of square by an hexagonal grid
                of hexagon. In this context:
                    `square_width` becomes the hexagon side length
                    `square_gap` becomes the gap between hexagon in all direction
                    `only_square` means only hexagon
                Defaults to False.
            layer: gds layer of the grid cover.
                Defaults to 1.
            datatype: gds datatype of the grid cover.
                Defaults to 1.
            name: gds name of the grid cover.
                Defaults to ''.
            color: gds color of the grid cover.
                Defaults to ''.
    """
    poly = offset(polygons=merge(polygons),
                  distance=-safety_margin,
                  join_first=True,
                  layer=layer,
                  datatype=datatype,
                  name=name,
                  color=color)

    if poly is not None:

        # First we merge the polygons and then we iterate over it
        # That way avoid to create grid in between polygons
        poly = merge(poly)

        resultPoly = PolygonSet()

        for p in poly.polygons:

            # skip empty polygon
            if (p==[[0,0]]).all():
                continue

            # Get the number of square in both direction to cover the polygon
            dx = np.ptp(p[:, 0])
            dy = np.ptp(p[:, 1])



            # Make an array that contain all coordinate of all squares at once
            # [(x0, y0),
            #  (x1, y1),
            #  (x2, y2),  that's the first square
            #  (x3, y3),
            #  .
            #  .
            #  .
            #  (xn-3, yn-3),
            #  (xn-2, yn-2), that's the n square where: n = nb_square_x * nb_square_y
            #  (xn-1, yn-1),
            #  (xn, yn)]
            #
            # To do so we:
            #   1. create an array of the square coordinate
            #   2. tile it to the number of need square
            #   3. add the square array to the translation coordinate



            # At that point, we pave the plan with the temp polygon array
            #   1. By default, the paving is a square paving
            #   2. If hexagonal_grid=True, we use a hexagon paving
            #      In this case `square_gap` is used as the hexagon side
            if hexagonal_grid:

                nb_square_x = ceil(dx/(square_width+square_gap)/2)
                # * 2 because the hex grid has two sub grid
                nb_square_y = ceil(dy/(square_width+square_gap)/2*2)

                # some trigo shortcut
                a = 0.5
                b = 0.8660254037844386

                # We keep track of the number of coordinate for the square
                split_nb = 6

                temp = np.tile(np.array([(0,                           0),
                                         (square_width,                0),
                                         (square_width+square_width*a, square_width*b),
                                         (square_width,                square_width*b*2),
                                         (0,                           square_width*b*2),
                                         (-square_width*a,             square_width*b)]
                                       , dtype=np.float64),
                                (nb_square_x*nb_square_y,1))

                # create the coordinates where the above hexagon should be repeated
                dx = np.tile(np.repeat(np.vstack(np.arange(0, nb_square_x, 1, dtype=np.float64)), split_nb, 0), (nb_square_y, 1))
                dy = np.repeat(np.repeat(np.vstack(np.arange(0, nb_square_y, 1, dtype=np.float64)), split_nb, 0), nb_square_x, 0)

                # Move the hex to the 1st subgrid
                dx *= (square_width+2*a*square_width+2*b*square_gap+square_width)
                dy *= (square_width*b*2+square_gap)
                temp2 = temp + np.hstack((dx, dy))

                # Move the hex to the second subgrid
                temp3 = temp + np.hstack((dx+square_width+b*square_gap+a*square_width,
                                dy-np.sqrt(3)/2*square_width -a*square_gap))

                # Make the total grid
                temp = np.concatenate((temp2, temp3))

            else:

                            # We keep track of the number of coordinate for the hexagon
                split_nb = 4

                nb_square_x = ceil(dx/(square_width+square_gap))
                nb_square_y = ceil(dy/(square_width+square_gap))

                temp = np.tile(np.array([(0,            0),
                                         (0,            square_width),
                                         (square_width, square_width),
                                         (square_width, 0)],
                                       dtype=np.float64)
                              , (nb_square_x*nb_square_y,1))

                # create the coordinates where the above square should be repeated
                dx = np.tile(np.repeat(np.vstack(np.arange(0, nb_square_x, 1, dtype=np.float64)), split_nb, 0), (nb_square_y, 1))
                dy = np.repeat(np.repeat(np.vstack(np.arange(0, nb_square_y, 1, dtype=np.float64)), split_nb, 0), nb_square_x, 0)

                temp += np.hstack((dx, dy))*(square_width+square_gap)


            # If user wants to add some noise to the square positions
            if noise!=0:
                t = 2*noise * np.random.random((int(temp.shape[0]/split_nb), 2)) - noise
                temp += np.repeat(t, split_nb, axis=0)

            # If use wants to center the whole grid
            if centered:
                temp += np.array([-(temp[:,0].max()+temp[:,0].min())/2, -(temp[:,1].max()+temp[:,1].min())/2])
                temp += np.array([(p[:,0].max()+p[:,0].min())/2, (p[:,1].max()+p[:,1].min())/2])
            else:
                temp += np.array([p[:,0].min(), p[:,1].min()])


            # Boolean operation
            polys = clipper.clip(np.split(temp, int(len(temp[:,0])/split_nb)),
                        [p],
                        'and',
                        1000)

            # if we still have some square left
            if len(polys) >0:
                resultPoly += PolygonSet(polys)
    else:
        resultPoly = PolygonSet()

    # If user wants to keep only entire squares
    if only_square:

        # Only check non empty PolygonSet
        if np.all(resultPoly.polygons[0]!=np.array([[0., 0.]])):
            temp = []
            for p in resultPoly.polygons:
                # We keep only quadrilateral
                if p.shape[0]==split_nb and p.shape[1]==2:
                    if np.isclose(np.sum(np.hypot( *(p - np.roll(p, -1, axis=0)).T )), split_nb*square_width, atol=1e-2):
                        temp.append(p)

            # if we still have some square left
            if len(temp)>0:
                resultPoly = PolygonSet(temp)
            else:
                resultPoly = PolygonSet()

    return resultPoly.change_layer(layer=layer,
                                   datatype=datatype,
                                   name=name,
                                   color=color)


def inverse_polarity(polygons: PolygonSet,
                     safety_marge:float=1e-3,
                     layer: int=1,
                     datatype: int=0,
                     name: str='',
                     color: str ='') -> Optional[PolygonSet]:
    """
    Inverse the polarity of the polygons.
    To do so a marge of 1nm is added to the 4 most extreme points.

    Parameters
    -----------
    safety_marge : float
        Marge added for the polarity inversion.
    """
    bounding_box = polygons.get_bounding_box()

    if bounding_box is None:
        warnings.warn("[pygdsdesign] Can't inverse_polarity a polygon with a null area.")
        return None
    a, b = bounding_box
    r = Rectangle((a[0]-safety_marge, a[1]-safety_marge),
                  (b[0]+safety_marge, b[1]+safety_marge))
    poly1 = r.polygons
    poly2 = polygons.polygons
    result = clipper.clip(poly1, poly2, "xor", 1 / 0.001)

    if len(result) == 0:
        warnings.warn("[pygdsdesign] The result of 'inverse_polarity' is None",stacklevel=4,)
        return None

    return PolygonSet(result, [layer]*len(result), [datatype]*len(result), [name]*len(result), [color]*len(result))


def inside(points: Coordinates,
           polygons: PolygonSet,
           short_circuit: Literal['any', 'all']='any',
           precision: float=0.001) -> Tuple[bool]:
    """
    Test whether each of the points is within the given set of polygons.

    Parameters
    ----------
    points : array-like[N][2] or sequence of array-like[N][2]
        Coordinates of the points to be tested or groups of points to be
        tested together.
    polygons : `PolygonSet`, `CellReference`, `CellArray` or iterable
        Polygons to be tested against.  If this is an iterable, each
        element must be a `PolygonSet`, `CellReference`, `CellArray`,
        or an array-like[N][2] of vertices of a polygon.
    short_circuit : {'any', 'all'}
        If `points` is a sequence of point groups, testing within each
        group will be short-circuited if any of the points in the group
        is inside ('any') or outside ('all') the polygons.  If `points`
        is simply a sequence of points, this parameter has no effect.
    precision : float
        Desired precision for rounding vertice coordinates.

    Returns
    -------
    out : tuple
        Tuple of booleans indicating if each of the points or point
        groups is inside the set of polygons.
    """
    polys = _gather_polys(polygons)
    pts: Tuple[List[np.ndarray]]|Coordinates

    if np.isscalar(points[0][0]):
        pts = (points,)
        sc = 0
    else:
        pts = points
        sc = 1 if short_circuit == "any" else -1

    return clipper.inside(pts, polys, sc, 1 / precision)


def copy(obj: Any,
         dx:float=0,
         dy:float=0) -> Any:
    """
    Create a copy of `obj` and translate it by (dx, dy).

    Parameters
    ----------
    obj : translatable object
        Object to be copied.
    dx : number
        Distance to move in the x-direction.
    dy : number
        Distance to move in the y-direction.


    Returns
    -------
    out : translatable object
        Translated copy of original `obj`

    Examples
    --------
    >>> rectangle = Rectangle((0, 0), (10, 20))
    >>> rectangle2 = copy(rectangle, 2,0)
    >>> myCell.add(rectangle)
    >>> myCell.add(rectangle2)
    """

    newObj = libcopy.deepcopy(obj)
    if dx != 0 or dy != 0:
        newObj.translate(dx, dy)
    return newObj


def select_polygon_per_layer(polygons: PolygonSet,
                             layer: int,
                             datatype: int=0,
                             merging: bool=False) -> PolygonSet:
    """
    Return a copy of the polygon(s) corresponding to the given layer and
    datatype.
    By default the datatype is 0.
    This method does not change the original Polygon or PolygonSet.

    Args:
        layer : Layer number.
        datatype: gds datatype of the grid cover.
            Defaults to 0.
        merging : If merging is True, merge before returning.
            Defaults to False.
    """

    # We create a numpy mask of the single layer we want to keep
    ls = np.array(polygons.layers)
    mask1 = ls==layer

    # We create a numpy mask of the single datatype we want to keep
    ds = np.array(polygons.datatypes)
    mask2 = ds==datatype

    # We combine both mask with an "and" boolean
    mask = mask1*mask2

    # We create an empty Polygon
    tot = PolygonSet()

    # We fill that polygon with the mask
    tot.polygons  = list(np.array(polygons.polygons, dtype=object)[mask])
    tot.layers    = list(ls[mask])
    tot.datatypes = list(ds[mask])
    tot.colors    = list(np.array(polygons.colors)[mask])
    tot.names     = list(np.array(polygons.names)[mask])

    if merging:
        return merge(tot)
    else:
        return tot


def select_polygon_per_name(polygons: PolygonSet,
                            name: int,
                            merging: bool=False) -> PolygonSet:
    """
    Return a copy of the polygon(s) corresponding to the given name
    This method does not change the original Polygon or PolygonSet.

    Args:
        layer : Layer number.
        datatype: gds datatype of the grid cover.
            Defaults to 0.
        merging : If merging is True, merge before returning.
            Defaults to False.
    """

    # We create a numpy mask of the single layer we want to keep
    ns = np.array(polygons.names)
    mask = ns==name

    # We create an empty Polygon
    tot = PolygonSet()

    # We fill that polygon with the mask
    tot.polygons  = list(np.array(polygons.polygons, dtype=object)[mask])
    tot.layers    = list(np.array(polygons.layers)[mask])
    tot.datatypes = list(np.array(polygons.datatypes)[mask])
    tot.colors    = list(np.array(polygons.colors)[mask])
    tot.names     = list(ns)

    if merging:
        return merge(tot)
    else:
        return tot



import numpy as np
import copy

from pygdsdesign import GdsLibrary, PolygonSet, Rectangle, Text, colors

offset_value = 0.250

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('TOP')

# Create empty polygon to which other polygon will be added
width  = 50
height = 10
gap_v  = 5
gap_h  = 10

tot = PolygonSet()
r1  = Rectangle((0, 0),
                (width, height))

for i, (name, color) in enumerate(colors.items()):

    # Others is not a material
    if name=='others':
        continue

    # Add a colored rectangle
    r2 = copy.copy(r1)
    r2.names = [name]
    r2.layers = [i]
    tot += r2.translate(0, tot.get_size()[1]+gap_v)

    # Add a name before the colored rectangle
    if name=='':
        name='default'
    t = Text(text=name, size=100, layer=i)
    t.scale(height/t.get_size()[1])
    tot += t.translate(-t.get_size()[0] - gap_h, -t.get_bounding_box()[0][1] + r2.get_bounding_box()[0][1])


# Add polygons to cell
cell.add(tot)

# Save GDS file
lib.export_gds('layers.gds', fill=False)



from functools import partial
import numpy as np
from copy import deepcopy
from typing import Callable, Tuple, Optional, Dict, Union
from typing_extensions import Literal
from scipy.integrate import quad
from scipy.interpolate import interp1d
from tqdm import tqdm

from pygdsdesign.transmission_lines.transmission_line import TransmissionLine
from pygdsdesign.polygons import Rectangle
from pygdsdesign.polygonSet import PolygonSet
from pygdsdesign.typing_local import Coordinate


class MicroStrip(TransmissionLine):

    def __init__(
        self,
        width: float,
        layer: int = 0,
        datatype: int = 0,
        name: str = "",
        color: str = "",
        ref: Optional[Coordinate] = None,
    ) -> None:
        """
        Microstrip allows to easily draw a continuous microstrip line.

        Parameters
        ----------
        width : float
            Width of the microstrip in um
            This width can be modified latter along the strip or smoothly
            by using tappered functions.
        layer : int
            Layer number of the microstrip.
        datatype : int
            Datatype number of the microstrip.
        """

        TransmissionLine.__init__(self, layer=layer,
                                        datatype=datatype,
                                        name=name,
                                        color=color,
                                        ref=ref)

        self._w = width


    @property
    def width(self):
        return self._w


    @width.setter
    def width(self, width:float):

        self._w = width


    @property
    def param_curve(self):
        return self._param_curve


    ###########################################################################
    #
    #                   Add polygons to the existing microstrip
    #
    ###########################################################################


    def add_line(self, x: float,
                       y: float) -> PolygonSet:
        """
        Add a piece of linear microstrip in the x or y direction .

        Parameters
        ----------
        x : float
            Length of the strip in the x direction in um.
        y : float
            Length of the strip in the y direction in um.
        """

        if x==0:
            if y>0:
                r = Rectangle((self.ref[0], self.ref[1]),
                              (self.ref[0]+self._w, self.ref[1]+y),
                              layer=self._layer,
                              datatype=self._datatype,
                              name=self._name,
                              color=self._color)
            else:
                r = Rectangle((self.ref[0], self.ref[1]),
                                    (self.ref[0]+self._w, self.ref[1]+y),
                                   layer=self._layer,
                                   datatype=self._datatype,
                                   name=self._name,
                                   color=self._color)

            self._add2param([self.ref[0]+self._w/2, self.ref[0]+self._w/2],
                            [self.ref[1], self.ref[1]+y],
                            [0, abs(y)])
        else:
            if x>0:
                r = Rectangle((self.ref[0], self.ref[1]),
                                    (self.ref[0]+x, self.ref[1]+self._w),
                                   layer=self._layer,
                                   datatype=self._datatype,
                                   name=self._name,
                                   color=self._color)
            else:
                r = Rectangle((self.ref[0], self.ref[1]),
                                    (self.ref[0]+x, self.ref[1]+self._w),
                                   layer=self._layer,
                                   datatype=self._datatype,
                                   name=self._name,
                                   color=self._color)

            self._add2param([self.ref[0], self.ref[0]+x],
                            [self.ref[1]+self._w/2, self.ref[1]+self._w/2],
                            [0, abs(x)])

        self._add(r)
        self.ref = [self.ref[0]+x, self.ref[1]+y]
        self.total_length += abs(x) + abs(y)

        return self


    ###########################################################################
    #
    #                       Add turn
    #
    ###########################################################################


    def add_turn(self, radius: float,
                       orientation: Literal['lb', 'l-b', 'left-bottom', 'leftbottom',
                                            'lt', 'l-t', 'left-top', 'lefttop'
                                            'rb', 'r-b', 'right-bottom', 'rightbottom'
                                            'rt', 'r-t', 'right-top', 'righttop'
                                            'tl', 't-l', 'top-left', 'topleft'
                                            'tr', 't-r', 'top-right', 'topright'
                                            'bl', 'b-l', 'bottom-left', 'bottomleft'
                                            'br', 'b-r', 'bottom-right', 'bottomright'
                                            'lt', 'l-t', 'lefttop', 'left-top'],
                       nb_points: int=50) -> PolygonSet:
        """
        Add a circulare turn to the strip.

        Parameters
        ----------
        radius : float
            Radius of the arc in um.
        orientation : str
            Orientation of the turn.
            The logic of the orientation naming is the following, the first term
            refers to your departure point while the second one refers to your
            arrival point, they are so the reverse to one another.
            Example: lt _|.
            Also the orientation can be writen in various way, for example all
            following are equivalent:
            'lb', 'l-b', 'left-bottom', 'leftbottom'
        nb_points : int (default=50)
            Number of point used in the polygon.
        """

        if orientation.lower() in ['lb', 'l-b', 'left-bottom', 'leftbottom']:
            theta = np.linspace(np.pi/2., 0, nb_points)
            x0 = self.ref[0]
            y0 = self.ref[1] - radius + self._w/2.

            self.ref = [self.ref[0] + radius - self._w/2., self.ref[1] - radius + self._w/2.]
        elif orientation.lower() in ['lt', 'l-t', 'left-top', 'lefttop']:
            theta = np.linspace(-np.pi/2., 0, nb_points)
            x0 = self.ref[0]
            y0 = self.ref[1] + radius + self._w/2.

            self.ref = [self.ref[0] + radius - self._w/2., self.ref[1] + radius + self._w/2.]
        elif orientation.lower() in ['rb', 'r-b', 'right-bottom', 'rightbottom']:

            theta = np.linspace(np.pi/2., np.pi, nb_points)
            x0 = self.ref[0]
            y0 = self.ref[1] - radius + self._w/2.

            self.ref = [self.ref[0] - radius - self._w/2., self.ref[1] - radius + self._w/2.]
        elif orientation.lower() in ['rt', 'r-t', 'right-top', 'righttop']:

            theta = np.linspace(-np.pi/2., -np.pi, nb_points)
            x0 = self.ref[0]
            y0 = self.ref[1] + radius + self._w/2.

            self.ref = [self.ref[0] - radius - self._w/2., self.ref[1] + radius + self._w/2.]
        elif orientation.lower() in ['tl', 't-l', 'top-left', 'topleft']:

            theta = np.linspace(0, -np.pi/2., nb_points)
            x0 = self.ref[0] - radius + self._w/2.
            y0 = self.ref[1]

            self.ref = [self.ref[0] - radius + self._w/2., self.ref[1] - radius - self._w/2.]
        elif orientation.lower() in ['tr', 't-r', 'top-right', 'topright']:

            theta = np.linspace(-np.pi, -np.pi/2., nb_points)
            x0 = self.ref[0] + radius + self._w/2.
            y0 = self.ref[1]

            self.ref = [self.ref[0] + radius + self._w/2., self.ref[1] - radius - self._w/2.]
        elif orientation.lower() in ['bl', 'b-l', 'bottom-left', 'bottomleft']:
            theta = np.linspace(0, np.pi/2., nb_points)
            x0 = self.ref[0] - radius + self._w/2.
            y0 = self.ref[1]

            self.ref = [self.ref[0] - radius + self._w/2., self.ref[1] + radius - self._w/2.]
        elif orientation.lower() in ['br', 'b-r', 'bottom-right', 'bottomright']:
            theta = np.linspace(np.pi, np.pi/2., nb_points)
            x0 = self.ref[0] + radius + self._w/2.
            y0 = self.ref[1]

            self.ref = [self.ref[0] + radius + self._w/2., self.ref[1] + radius - self._w/2.]
        else:
            raise ValueError('Your orientation "'+orientation+'" is incorrect.')

        x = np.concatenate(((radius-self._w/2.)*np.cos(theta), (radius + self._w/2.)*np.cos(theta[::-1])))
        y = np.concatenate(((radius-self._w/2.)*np.sin(theta), (radius + self._w/2.)*np.sin(theta[::-1])))

        p = np.vstack((x0+x, y0+y)).T

        self._add(PolygonSet(polygons=[p],
                             layers=[self._layer],
                             datatypes=[self._datatype],
                             names=[self._name],
                             colors=[self._color]))

        self.total_length += radius*np.pi/2.

        self._add2param(x0+radius*np.cos(theta),
                        y0+radius*np.sin(theta),
                        np.linspace(0, 1, nb_points)*radius*np.pi/2)

        return self


    def add_fresnel_turn(self, radius: float,
                               orientation: Literal['lb', 'l-b', 'left-bottom', 'leftbottom',
                                                    'lt', 'l-t', 'left-top', 'lefttop'
                                                    'rb', 'r-b', 'right-bottom', 'rightbottom'
                                                    'rt', 'r-t', 'right-top', 'righttop'
                                                    'tl', 't-l', 'top-left', 'topleft'
                                                    'tr', 't-r', 'top-right', 'topright'
                                                    'bl', 'b-l', 'bottom-left', 'bottomleft'
                                                    'br', 'b-r', 'bottom-right', 'bottomright'
                                                    'lt', 'l-t', 'lefttop', 'left-top'],
                               nb_points: int=51) -> PolygonSet:
        """
        Return the parametric Fresnel curve along the length t.
        The curve is doubled, mirrored, joint and normalized to a given radius

        Parameters
        ----------
        radius : float
            Radius of the arc in um.
        orientation : str
            Orientation of the turn.
            The logic of the orientation naming is the following, the first term
            refers to your departure point while the second one refers to your
            arrival point, they are so the reverse to one another.
            Example: lt _|.
            Also the orientation can be writen in various way, for example all
            following are equivalent:
            'lb', 'l-b', 'left-bottom', 'leftbottom'
        nb_points : int (default=51)
            Number of point used in the polygon.
            Must be an odd number
        """



        # All the computations are stored in self.calc to cache the result
        dx, dy, t, x, y, f_dx, f_dy, length = self.calc(radius, nb_points)

        # Prepare functions for the parametric generation
        # Initialize func and dfunc with some values to reuse them later on
        partial_func = partial(self.func, x=x,y=y)
        partial_dfunc = partial(self.dfunc, dx=dx,dy=dy,f_dx=f_dx,f_dy=f_dy)

        # Get the polygon corresponding to a Fresnel turn
        # Add_length is set to False, because of its computational time.
        # We do the computation here and save it for later
        p = self.add_parametric_curve(partial_func, partial_dfunc, t,
                args=None,
                add_polygon=False,
                add_length=False)


        # Add the length of the parametric curve only if asked (default)
        if hasattr(self, '_fresnel_lengths'):
            if radius not in self._fresnel_lengths.keys():
                compute_length = True
            else:
                compute_length = False
        else:
            self._fresnel_lengths: Dict[float, float]= {}
            compute_length = True

        if compute_length:
            self._fresnel_lengths[radius] = length

        self.total_length += self._fresnel_lengths[radius]
        t = np.linspace(0, 1, nb_points)*self._fresnel_lengths[radius]

        # We transform the polygon and move the ref depending on the orientation
        if orientation.lower() in ['lt', 'l-t', 'lefttop', 'left-top']:
            self._add2param(self.ref[0]+x[::-1], self.ref[1]+y[::-1], t)
            self.ref = p.polygons[0][int(nb_points-1)]
            self._add(p)
        elif orientation.lower() in ['lb', 'l-b', 'leftbottom', 'left-bottom']:
            p.flip('x')
            p.translate(0, -radius+self._w/2.)
            self._add2param(self.ref[0]+x[::-1], self.ref[1]-y[::-1]+self._w, t)
            self.ref = [p.polygons[0][nb_points][0]-self._w, p.polygons[0][nb_points][1]]
            self._add(p)
        elif orientation.lower() in ['br', 'b-r', 'bottomright', 'bottom-right']:
            p.flip('x')
            p.flip('y')
            self._add2param(self.ref[0]+y[::-1], self.ref[1]+x[::-1], t)
            self.ref = p.polygons[0][0]
            self._add(p)
        elif orientation.lower() in ['bl', 'b-l', 'bottomright', 'bottom-right']:
            p.rotate(np.pi/2., p.polygons[0][0]+self._w/2)
            p.translate(-self._w, -self._w)
            self._add2param(self.ref[0]-y[::-1]+self._w, self.ref[1]+x[::-1], t)
            self.ref = [p.polygons[0][nb_points][0], p.polygons[0][nb_points][1]-self._w]
            self._add(p)
        elif orientation.lower() in ['tr', 't-r', 'topright', 'top-right']:
            p.rotate(-np.pi/2., p.polygons[0][0]+self._w/2)
            p.translate(self._w, -2*self._w)
            self._add2param(self.ref[0]+y[::-1], self.ref[1]-x[::-1], t)
            self.ref = [p.polygons[0][nb_points][0], p.polygons[0][nb_points][1]]
            self._add(p)
        elif orientation.lower() in ['tl', 't-l', 'topleft', 'top-left']:
            p.rotate(-np.pi/2., p.polygons[0][0]+self._w/2)
            p.flip('y')
            p.translate(-radius+3*self._w/2.,-2*self._w)
            self._add2param(self.ref[0]-y[::-1]+self._w, self.ref[1]-x[::-1], t)
            self.ref = [p.polygons[0][nb_points][0], p.polygons[0][nb_points][1]]
            self._add(p)
        elif orientation.lower() in ['rb', 'r-b', 'rightbottom', 'right-bottom']:
            p.flip('x').flip('y')
            p.translate(-radius-self._w/2, -radius+self._w/2)
            self._add2param(self.ref[0]-x[::-1], self.ref[1]-y[::-1]+self._w, t)
            self.ref = [p.polygons[0][nb_points][0], p.polygons[0][nb_points][1]]
            self._add(p)
        elif orientation.lower() in ['rt', 'r-t', 'righttop', 'right-top']:
            p.flip('y')
            p.translate(-radius-self._w/2, 0)
            self._add2param(self.ref[0]-x[::-1], self.ref[1]+y[::-1], t)
            self.ref = [p.polygons[0][nb_points][0], p.polygons[0][nb_points][1]]
            self._add(p)
        else:
            raise ValueError('Your orientation "'+orientation+'" is incorrect.')
        return self


    ###########################################################################
    #
    #                               Tapers
    #
    ###########################################################################


    def add_taper(self, x: float,
                        y: float,
                        new_width: float) -> PolygonSet:
        """
        Add linear taper between the current and the new width.

        Parameters
        ----------
        x : float
            Length of the taper in the x direction in um.
        y : float
            Length of the taper in the y direction in um.
        new_width : float
            New width of the microstrip in um.
        """

        if x == 0 and y > 0:
            p = [(self.ref[0], self.ref[1]),
                 (self.ref[0]-new_width/2.+self._w/2., self.ref[1]+y),
                 (self.ref[0]+new_width/2.+self._w/2., self.ref[1]+y),
                 (self.ref[0]+self._w, self.ref[1])]

            self._add2param([self.ref[0]+self._w/2, self.ref[0]+self._w/2],
                            [self.ref[1], self.ref[1]+y],
                            [0, abs(y)])
            self.ref = [self.ref[0]-new_width/2.+self._w/2., self.ref[1]+y]
        elif x == 0 and y < 0:
            p = [(self.ref[0], self.ref[1]),
                 (self.ref[0]-new_width/2. +self._w/2., self.ref[1]+y),
                 (self.ref[0]+new_width/2. +self._w/2., self.ref[1]+y),
                 (self.ref[0]+self._w, self.ref[1])]

            self._add2param([self.ref[0]+self._w/2, self.ref[0]+self._w/2],
                            [self.ref[1], self.ref[1]+y],
                            [0, abs(y)])
            self.ref = [self.ref[0]-new_width/2.+self._w/2., self.ref[1]+y]
        elif x > 0 and y==0:

            p = [(self.ref[0], self.ref[1]),
                 (self.ref[0]+x, self.ref[1] - new_width/2.+self._w/2.),
                 (self.ref[0]+x, self.ref[1] + new_width/2.+self._w/2.),
                 (self.ref[0], self.ref[1]+self._w)]

            self._add2param([self.ref[0], self.ref[0]+x],
                            [self.ref[1]+self._w/2, self.ref[1]+self._w/2],
                            [0, abs(x)])
            self.ref = [self.ref[0]+x, self.ref[1]-new_width/2.+self._w/2.]
        elif x<0 and y==0:

            p = [(self.ref[0], self.ref[1]),
                 (self.ref[0]+x, self.ref[1] - new_width/2.+self._w/2.),
                 (self.ref[0]+x, self.ref[1] + new_width/2.+self._w/2.),
                 (self.ref[0], self.ref[1]+self._w)]

            self._add2param([self.ref[0], self.ref[0]+x],
                            [self.ref[1]+self._w/2, self.ref[1]+self._w/2],
                            [0, abs(x)])
            self.ref = [self.ref[0]+x, self.ref[1]-new_width/2.+self._w/2.]

        self._add(PolygonSet(polygons=[p],
                             layers=[self._layer],
                             datatypes=[self._datatype],
                             names=[self._name],
                             colors=[self._color]))

        self.total_length += abs(x) + abs(y)
        self._w = new_width

        return self


    def add_taper_cosec(self, x: float,
                              y: float,
                              new_width: float,
                              smoothness: float=1,
                              nb_points: int=50) -> PolygonSet:
        """
        Add smooth taper between the current and the new width.
        The equation of the smooth curve is based on the cosec function

        Parameters
        ----------
        x : float
            Length of the taper in the x direction in um.
        y : float
            Length of the taper in the y direction in um.
        new_width : float
            New width of the microstrip in um.
        smoothness : float (default 1)
            Slop of the taper, smaller number implying sharper transition.
        nb_points : int (default=50)
            Number of point used in the polygon.
        """

        # Calculate curve ensuring it goes from y=0 to y=1
        tx  = np.linspace(0., abs(x)+abs(y), nb_points)
        ty = 1./((1.+np.exp((tx-(abs(x)+abs(y))/2.)/smoothness))/np.exp((tx-(abs(x)+abs(y))/2.)/smoothness))
        ty -= ty.min()
        ty /= ty.max()

        # Scale curve to proper microstrip width and tapering
        ty = self._w/2. + (new_width-self._w)*ty/2.

        # Build ref polygon
        tx = np.concatenate((tx, tx[::-1]))
        ty = np.concatenate((ty, ty[::-1]*-1))
        p = PolygonSet(polygons=[np.vstack((tx, ty)).T],
                       layers=[self._layer],
                       datatypes=[self._datatype],
                       names=[self._name],
                       colors=[self._color])

        # Transform polygon depending on orientation
        if y==0 and x>0:
            p.translate(self.ref[0], self.ref[1]+self._w/2.)
            self._add2param([self.ref[0], self.ref[0]+x],
                            [self.ref[1]+self._w/2, self.ref[1]+self._w/2],
                            [0, abs(x)])
            self.ref = [self.ref[0]+x, self.ref[1]-new_width/2.+self._w/2.]
        elif y==0 and x<0:
            p.polygons[0]=np.vstack((p.polygons[0][:,0]*-1, p.polygons[0][:,1])).T
            p.translate(self.ref[0], self.ref[1]+self._w/2.)
            self._add2param([self.ref[0], self.ref[0]+x],
                            [self.ref[1]+self._w/2, self.ref[1]+self._w/2],
                            [0, abs(x)])
            self.ref = [self.ref[0]+x, self.ref[1]-new_width/2.+self._w/2.]
        elif x==0 and y>0:
            p.translate(0, -self._w/2.)
            p.rotate(np.pi/2., center=(0, 0))
            p.translate(self.ref[0], self.ref[1])
            self._add2param([self.ref[0]+self._w/2, self.ref[0]+self._w/2],
                            [self.ref[1], self.ref[1]+y],
                            [0, abs(y)])
            self.ref = [self.ref[0]-new_width/2.+self._w/2., self.ref[1]+y]
        elif x==0 and y<0:
            p.rotate(-np.pi/2., center=(0, 0))
            p.translate(self._w/2., 0)
            p.translate(self.ref[0], self.ref[1])
            self._add2param([self.ref[0]+self._w/2, self.ref[0]+self._w/2],
                            [self.ref[1], self.ref[1]+y],
                            [0, abs(y)])
            self.ref = [self.ref[0]-new_width/2.+self._w/2., self.ref[1]+y]

        self._add(p)

        self.total_length += abs(x) + abs(y)
        self._w = new_width

        return self


    ###########################################################################
    #
    #                   Generic parametric curve
    #
    ###########################################################################


    def add_parametric_curve(self, f: Callable[..., Tuple[np.ndarray, np.ndarray]],
                                  df: Callable[..., Tuple[np.ndarray, np.ndarray]],
                                   t: np.ndarray,
                                args: Optional[Tuple[Optional[float], ...]]=None,
                                add_polygon: bool=True,
                                add_length: bool=True) -> PolygonSet:
        """
        Create a microstrip line following the parametric equation f and its
        derivative df along the length t.
        In order to return the curve length correctly, the derivative df of f
        must be correct, its absolute amplitude must be correct.

        Parameters
        ---------
        f : func
            Function calculating the parametric equation.
            Must be of the type f(t, args) and return a tuple of coordinate
            (x, y).
        df : func
            Function calculating the derivative of the parametric equation.
            Must be of the type df(t, args) and return a tuple of coordinate
            (dx, dy).
        t : np.ndarray
            Array of the length of the parametric curve.
            Also determine the number of point of the total polygon.
            Must not necessarily starts at 0.
        args : variable arguments (default None)
            Argument passed to f and df
        """
        if args is None:
            args = (None, )

        dx1, dy1 = df(t, args)

        n = np.hypot(dx1, dy1)

        dx1, dy1 = dx1/n, dy1/n

        x1_, y1_ = f(t, args)
        theta1 = np.angle(dx1+1j*dy1)-np.pi/2.
        x1, y1 = x1_+np.cos(theta1)*self._w/2., y1_+np.sin(theta1)*self._w/2.


        dx2, dy2 = df(t[::-1], args)
        n = np.hypot(dx2, dy2)
        dx2, dy2 = dx2/n, dy2/n
        x2_, y2_ = f(t[::-1], args)
        theta2 = np.angle(dx2+1j*dy2)+np.pi/2.
        x2, y2 = x2_+np.cos(theta2)*self._w/2., y2_+np.sin(theta2)*self._w/2.

        x, y = np.concatenate((x1, x2)), np.concatenate((y1, y2))
        p = np.vstack((x, y)).T

        poly = PolygonSet(polygons=[p],
                         layers=[self._layer],
                         datatypes=[self._datatype],
                         names=[self._name],
                         colors=[self._color]).translate(self.ref[0], self.ref[1])

        # Calculate curve length
        def func(t, args):
            dx1, dy1 = df(t, args)
            return np.hypot(dx1, dy1)

        # Add the length of the parametric curve only if asked (default)
        if add_length:
            self.total_length += quad(func, t[0], t[-1], args=(args,))[0]

        # Add polygon only if asked (default)
        if add_polygon:
            self._add2param(self.ref[0]+x1_,
                            self.ref[1]+y1_,
                            np.linspace(0, 1, len(t))*quad(func, t[0], t[-1], args=(args,))[0])
            self.ref = poly.polygons[0][int(len(poly.polygons[0])/2)]
            self._add(poly)

            return self
        else:
            return poly


    ###########################################################################
    #
    #                   Added structures following the strip
    #
    ###########################################################################


    def add_and_follow(self, structure: Union[PolygonSet, PolygonSet],
                             pitch: float=100.,
                             start_offset: float=0.,
                             stop_offset: float=0.,
                             distance: float=50.,
                             angle: float=0.,
                             noise: float=0.,
                             nb_point_linearization: int=1000,
                             add_polygon: bool=True,
                             verbose: bool=False) -> PolygonSet:
        """
        Add structures along the line. The structures follow the curve and are
        rotating accordingly.

        Args:
            structure : Structure to be added along the path.
            pitch: pitch between two adjacent structures in um.
                Defaults to 100um.
            start_offset: Offset between the beginning of the microstrip and the
                first added structure in um.
                Defaults to 0um.
            stop_offset: Offset between the end of the microstrip and the last
                added structure in um.
                Defaults to 0um.
            distance: distance between the center of the microstrip and the
                center of the added structure in um. This distance is refered to
                the orthogonal vector of the line. A positive distance shift
                structures on the "right" of the microstrip.
                Defaults to 50um.
            angle: angle used to rotate the structure in radian
                Defaults to 0rad.
            noise: Nornal noise added to the pitch of the added structures in um.
                Usefull to avoid a perfect distance between the added structures
                and spurious resonant effect.
                Defaults to 0um.
            nb_point_linearization: Internally the method linearizes the
                parametric curve describing the shape of the microstrip. This
                linearization requires a large number of points in order to
                obtain a precise estimation of the parametric curve.
                Defaults to 1000.
            add_polygon: If True add the added structures to the microstrip.
                If False, return the added structures as independent polygons.
                Defaults to True.
            verbose: If true, display a progress bar in the terminal. Usefull
                for time consuming rendering.
                Deafault to False.
        """

        # Center the structure
        structure.center()

        # First you need to linearize the parametric coordinate
        t = np.linspace(start_offset, self.param_curve[:,2][-1], nb_point_linearization)
        x = np.interp(t, self.param_curve[:,2], self.param_curve[:,0])
        y = np.interp(t, self.param_curve[:,2], self.param_curve[:,1])


        # From the linearized coordinates, we create coordinate functions
        f_x = interp1d(t, x)
        f_y = interp1d(t, y)

        f_dx = interp1d(t, np.gradient(x, t))
        f_dy = interp1d(t, np.gradient(y, t))

        f_theta = lambda t: np.angle(f_dx(t)+1j*f_dy(t))

        # For every pitch along the parametric curve, we add the given structure
        ts = np.linspace(start_offset,
                         self.param_curve[:,2][-1]-stop_offset,
                         int((self.param_curve[:,2][-1]-stop_offset)/pitch))

        # We add normal noise to the computed position
        # By default, noise is 0, adding no noise
        # If noise is not 0, we take care of out of range value with masks
        ts += np.random.normal(0, noise, len(ts))
        ts[ts<start_offset] = start_offset
        ts[ts>self.param_curve[:,2][-1]-stop_offset] = self.param_curve[:,2][-1]-stop_offset

        if verbose:
            it = tqdm
        else:
            it = lambda t: t

        poly = PolygonSet()
        for t in it(ts):

            # Compute the coordinate and angle of the curve at the distance t
            x1, y1, theta = f_x(t), f_y(t), f_theta(t)

            # With the angle, we compute the added coordinates giving the
            # wanted distance
            # The added -pi/2 is to give the distance in reference to the orthogonal
            # direction in respect to the parametric curve.
            x1 += np.cos(theta-np.pi/2)*distance
            y1 += np.sin(theta-np.pi/2)*distance

            # Add the structure at (order matters):
            # 1. the proper angle
            # 2. the proper coordinate
            poly += deepcopy(structure)\
                    .rotate(theta+angle, (0, 0))\
                    .translate(x1, y1)

        # Add polygon only if asked (default)
        if add_polygon:
            self += poly
            return self
        else:
            return poly



import numpy as np
import time

from pygdsdesign import GdsLibrary, Text, CPW


start = time.time()

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('FIRST')

# Test base
start_base = time.time()
c = CPW(width=2, gap=3, layer=6)
c.add_line(0, 15)
c.add_line(-15, 0)
c.add_line(0, -15)

c.add_line(15, 0)
c.add_turn(15, 'lb')
c.add_line(0, -15)
c.add_turn(15, 'tl')
c.add_line(-15, 0)
c.add_line(-20, 0)
c.add_turn(15, 'rt')
c.add_line(0, 15)
c.add_turn(15, 'bl')
c.add_line(-15, 0)
c.add_turn(15, 'rb')
c.add_line(0, -75)
c.add_turn(15, 'tr')
c.add_line(10, 0)
c.add_turn(15, 'lt')
c.add_line(0, 15)
c.add_turn(15, 'br')
c.add_line(15, 0)
c.add_turn(15, 'lb')
c.add_line(0, -75)
c.add_turn(15, 'tr')

c.add_line(5, 0)
c.add_taper(20, 0, 10, 23)
c.add_line(5, 0)
c.add_taper(20, 0, 2, 5)
c.add_turn(15, 'lb')
c.add_line(0, -5)
c.add_taper(0, -20, 5, 11)
c.add_line(0, -5)
c.add_taper(0, -20, 9, 2)
c.add_line(0, -5)
c.add_turn(15, 'tl')
c.add_line(-5, 0)
c.add_taper(-20, 0, 2, 5)
c.add_line(-5, 0)
c.add_taper(-20, 0, 9, 17)
c.add_line(-5, 0)
c.add_turn(15, 'rt')
c.add_line(0, 5)
c.add_taper(0, 10, 2, 3)
c.add_turn(15, 'bl')
c.add_line(-500, 0)
c.add_turn(15, 'rt')
c.add_line(0, 500)
c.add_turn(15, 'br')
# c.add_end(3, 0)
c.add_turn(15, 'lt')
# c.add_end(0, 3)
c.add_turn(15, 'bl')
# c.add_end(-3, 0)
c.add_turn(15, 'rb')
c.add_end(0, -3)

# c.inverse_polarity()
bounding_polygon = c.bounding_polygon.change_layer(10)
c+=bounding_polygon
cell.add(c)

stop_base = time.time()
print('Test base: {:.3f}s'.format(stop_base-start_base))

# Test circular turn, should form a close loop
start_circular = time.time()
m = CPW(2, 3, layer=2, name='loop')
m.add_line(0, 10)
m.add_turn(50, 'br')
m.add_line(10, 0)

m.add_turn(50, 'lt')
m.add_line(0, 10)
m.add_turn(50, 'br')
m.add_line(10, 0)
m.add_turn(50, 'lb')
m.add_line(0, -10)
m.add_turn(50, 'tr')
m.add_line(10, 0)
m.add_turn(50, 'lb')
m.add_line(0, -10)
m.add_turn(50, 'tl')
m.add_line(-10, 0)
m.add_turn(50, 'rb')
m.add_line(0, -10)
m.add_turn(50, 'tl')
m.add_line(-10, 0)
m.add_turn(50, 'rt')
m.add_line(0, 10)
m.add_turn(50, 'bl')
m.add_line(-10, 0)
m.add_turn(50, 'rt')

m.center()
m += Text('Circular loop', 50, (0, 0), layer=2, name='loop').center().translate(0, 250)

bounding_polygon = m.bounding_polygon.change_layer(10)
m+=bounding_polygon
cell.add(m.translate(-600, 600))

stop_circular = time.time()
print('Test circular: {:.3f}s'.format(stop_circular-start_circular))






# Test serpentine Circular
start_serpentine = time.time()
m = CPW(2, 3, layer=3, name='serpentine')

m.add_line(250, 0)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='right')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='right')
m.add_line(0, 250)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='top')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='top')
m.add_line(-500, 0)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='left')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='left')
m.add_line(0, -500)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='bottom')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='bottom')

m.center()
m += Text('Test serpentine circular', 50, (0, 0), layer=3, name='serpentine').center().translate(0, 600)
bounding_polygon = m.bounding_polygon.change_layer(10)
m+=bounding_polygon
cell.add(m.translate(1500, 1500))

stop_serpentine = time.time()
print('Test serpentine circular: {:.3f}s'.format(stop_serpentine-start_serpentine))




# Test serpentine Fresnel
start_serpentine = time.time()
m = CPW(2, 3, layer=3, name='serpentine')

m.add_line(250, 0)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='right', turn='fresnel')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='right', turn='fresnel')
m.add_line(0, 250)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='top', turn='fresnel')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='top', turn='fresnel')
m.add_line(-500, 0)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='left', turn='fresnel')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='left', turn='fresnel')
m.add_line(0, -500)
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='right', orientation='bottom', turn='fresnel')
m.add_serpentine(total_length=1000, nb_turn=4, spacing=35, starting='left', orientation='bottom', turn='fresnel')

m.center()
m += Text('Test serpentine fresnel', 50, (0, 0), layer=3, name='serpentine').center().translate(0, 600)

bounding_polygon = m.bounding_polygon.change_layer(10)
m+=bounding_polygon
cell.add(m.translate(0, 1500))

stop_serpentine = time.time()
print('Test serpentine fresnel: {:.3f}s'.format(stop_serpentine-start_serpentine))



# Test spiral
start_spiral = time.time()
def spiral(u, args):

    r = args[1]*2.*args[2]*u + args[0]/2.
    theta = args[1]*2.*np.pi*u
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return x, y

def dspiral_dt(u, args):

    r = args[1]*2.*args[2]*u
    theta = args[1]*2.*np.pi*u
    dx_dt = -args[1]*2.*np.pi*np.sin(theta)*r + np.cos(theta)*args[1]*2.*args[2]
    dy_dt =  args[1]*2.*np.pi*np.cos(theta)*r + np.sin(theta)*args[1]*2.*args[2]
    return dx_dt, dy_dt


t = np.linspace(0, 1, 10000)
inner_diameter = 10
nb_turn = 40
spiral_spacing = 10.
spiral_width = 2.

m = CPW(2, 3, layer=4, name='spiral')
m.add_parametric_curve(spiral, dspiral_dt, t,
                                args=(inner_diameter, nb_turn, spiral_spacing))
m.add_line(0, 100)

m.center()
m += Text('Test spiral', 50, (0, 0), layer=4, name='spiral').center().translate(0, 900)
bounding_polygon = m.bounding_polygon.change_layer(10)
m+=bounding_polygon
cell.add(m.translate(500, -1000))

stop_spiral = time.time()
print('Test spiral: {:.3f}s'.format(stop_spiral-start_spiral))





# Test Fresnel turn, should form a close loop
start_fresnel = time.time()
m = CPW(4, 5, layer=5, name='Fresnel')
m.add_line(0, 10)
m.add_turn(50, 'br')
m.add_line(10, 0)

m.add_fresnel_turn(50, 'lt', nb_points=51)
m.add_line(0, 10)
m.add_fresnel_turn(50, 'br')
m.add_line(10, 0)
m.add_fresnel_turn(50, 'lb')
m.add_line(0, -10)
m.add_fresnel_turn(50, 'tr')
m.add_line(10, 0)
m.add_fresnel_turn(50, 'lb')
m.add_line(0, -10)
m.add_fresnel_turn(50, 'tl')
m.add_line(-10, 0)
m.add_fresnel_turn(50, 'rb')
m.add_line(0, -10)
m.add_fresnel_turn(50, 'tl')
m.add_line(-10, 0)
m.add_fresnel_turn(50, 'rt')
m.add_line(0, 10)
m.add_fresnel_turn(50, 'bl')
m.add_line(-10, 0)
m.add_turn(50, 'rt')

m.center()
m += Text('Fresnel loop', 50, (0, 0), layer=5, name='Fresnel').center().translate(0, 250)
bounding_polygon = m.bounding_polygon.change_layer(10)
m+=bounding_polygon
cell.add(m.translate(500, 500))

stop_fresnel = time.time()
print('Test fresnel: {:.3f}s'.format(stop_fresnel-start_fresnel))


lib.export_gds('cpw.gds')

stop = time.time()

print('Generation time:', stop-start, ' s')


import copy

from pygdsdesign import GdsLibrary, PolygonSet, Rectangle, Text, MicroStrip

# Example of a microstrip transmission line where a multiple of identical
# resonator are coupled to in order to create a stop band


# Chip size of 10mm x 7mm10
chip_dx = 10000
chip_dy = 10000

# The GDSII file is called a library, which contains multiple cells.
lib = GdsLibrary()

# Geometry must be placed in cells.
cell = lib.new_cell('TOP')

# Create empty polygon to which other polygon will be added
tot = PolygonSet()

# Chip mark
p = Rectangle((0,0), (200, 10), layer=0).center()
p += Rectangle((0,0), (10, 200), layer=0).center()
dx_d, dy_d = p.get_size()
tot += p
tot += copy.deepcopy(p).translate(chip_dx, 0)
tot += copy.deepcopy(p).translate(chip_dx, -chip_dy)
tot += copy.copy(p).translate(0., -chip_dy)


# Default resonator
r = MicroStrip(width=2, layer=0)
r.add_line(150, 0)
r.add_fresnel_turn(25, 'lb')
r.add_fresnel_turn(25, 'tl')
r.add_line(-75, 0)
r.add_fresnel_turn(10, 'rb')
r.add_serpentine(total_length=1500,
                 nb_turn=10,
                 spacing=20,
                 orientation='bottom',
                 starting='left',
                 turn='fresnel',
                 )

# Feedline
pad_width = 500
pad_length = 500
offset_length = 500
taper_length = 300
turn_radius = 50
nb_turn = 8
spacing = 700
end_line = 100

m = MicroStrip(width=pad_width, layer=0)
m.add_line(pad_length, 0)
m.add_taper(taper_length, 0, 4)
m.add_line(offset_length, 0)
m.add_serpentine(total_length=50000,
                 nb_turn=nb_turn,
                 spacing=spacing,
                 orientation='right',
                 starting='left',
                 )
m.add_line(offset_length, 0)
m.add_taper(taper_length, 0, pad_width)
m.add_line(pad_length, 0)
m.add_and_follow(structure=r,
                 pitch=1000,
                 start_offset=pad_length+offset_length,
                 stop_offset=pad_length+offset_length,
                 noise=100,
                 distance=144.5)

tot += m.center().translate(chip_dx/2, -chip_dy/2)

# Chip label
t = Text('R0C0', 250)
dx_t, dy_t = t.get_size()
tot += t.translate(-dx_t+chip_dx-250, -dy_t-250)


# Center the design to the chip and the chip to the (0,0) of the cell
tot.center()
tot += p.center()

# Add polygons to cell
cell.add(tot)

# Save GDS file
lib.write_gds('microstrip_resonator.gds')
